import argparse
from copy import copy

import configargparse
import torch
from aislib.misc_utils import get_logger

from eir.train_utils.optimizers import get_base_optimizers_dict

logger = get_logger(name=__name__)


def get_default_cl_args():
    parser = get_train_argument_parser()
    cl_args = parser.parse_args()
    cl_args = modify_train_arguments(cl_args=cl_args)

    return cl_args


def get_train_argument_parser() -> configargparse.ArgumentParser:

    parser_ = configargparse.ArgumentParser(
        config_file_parser_class=configargparse.YAMLConfigFileParser
    )

    parser_.add_argument(
        "--config_file",
        is_config_file=True,
        required=False,
        help="path to .yaml config file if using one.",
    )

    parser_.add_argument(
        "--n_epochs", type=int, default=5, help="Number of epochs of training."
    )
    parser_.add_argument(
        "--batch_size", type=int, default=64, help="Size of the batches."
    )

    parser_.add_argument(
        "--dataloader_workers",
        type=int,
        default=8,
        help="Number of workers for multi-process training and validation data "
        "loading.",
    )
    parser_.add_argument(
        "--lr", type=float, default=1e-3, help="Base learning rate for optimizer."
    )

    parser_.add_argument(
        "--find_lr",
        action="store_true",
        help="Whether to perform a range test of different learning rates, with "
        "the lower limit being what is passed in for the --lr flag. "
        "Produces a plot and exits with status 0 before training if this flag "
        "is active.",
    )

    parser_.add_argument(
        "--lr_schedule",
        type=str,
        default="same",
        choices=["cycle", "plateau", "same", "cosine"],
        help="Whether to use cyclical, cosine or reduce on plateau learning rate "
        "schedule. Otherwise keeps same learning rate.",
    )

    parser_.add_argument(
        "--lr_plateau_patience",
        type=int,
        default=10,
        help="Number of validation performance steps without improvement over "
        "best performance before reducing LR (only relevant when --lr_schedule is "
        "'plateau'.",
    )

    parser_.add_argument(
        "--lr_plateau_factor",
        type=float,
        default=0.1,
        help="Factor to reduce LR when running with plateau schedule.",
    )

    parser_.add_argument(
        "--early_stopping_patience",
        type=int,
        default=None,
        help="Number of validation performance steps without improvement over "
        "best performance before terminating run.",
    )

    parser_.add_argument(
        "--early_stopping_buffer",
        type=int,
        default=None,
        help="Number of iterations to run before activation early stopping checks, "
        "useful if networks take a while to 'kick into gear'.",
    )

    parser_.add_argument(
        "--warmup_steps", type=str, default=0, help="How many steps to use in warmup."
    )

    parser_.add_argument(
        "--lr_lb",
        type=float,
        default=0.0,
        help="Lower bound for learning rate when using LR scheduling.",
    )

    parser_.add_argument(
        "--optimizer",
        type=str,
        choices=_get_optimizer_cl_arg_choices(),
        default="adam",
        help="What optimizer to use.",
    )

    parser_.add_argument(
        "--b1",
        type=float,
        default=0.9,
        help="Decay of first order momentum of gradient for relevant optimizers.",
    )

    parser_.add_argument(
        "--b2",
        type=float,
        default=0.999,
        help="Decay of second order momentum of gradient for relevant optimizers.",
    )
    parser_.add_argument("--wd", type=float, default=0.00, help="Weight decay.")

    parser_.add_argument(
        "--l1", type=float, default=0.00, help="L1 regularization for chosen layer."
    )

    parser_.add_argument(
        "--fc_repr_dim",
        type=int,
        default=512,
        help="Dimensionality of first fc layer in the network, this is the last shared"
        "when running multi task training, first fc layer after convolutions when"
        "running cnn model, and first fc layer when running mlp.",
    )

    parser_.add_argument(
        "--fc_task_dim",
        type=int,
        default=128,
        help="Dimensionality of (a) specific task branches in multi task setting, (b)"
        "successive fc layers in cnn model after the first fc after the "
        "convolutions and (c) successive fc layers in mlp after first fc layer.",
    )

    parser_.add_argument(
        "--mg_num_experts",
        type=int,
        default=8,
        help="Number of experts to use when using the MGMoE fusion model.",
    )

    parser_.add_argument(
        "--split_mlp_num_splits",
        type=int,
        default=50,
        help="Number of splits in split MLP layer. Warning: Will be deprecated.",
    )

    parser_.add_argument(
        "--model_type",
        type=str,
        default="cnn",
        choices=["cnn", "mlp", "mlp-split", "genome-local-net", "linear"],
        help="Model type for omics model.",
    )

    parser_.add_argument(
        "--fusion_model_type",
        type=str,
        default="default",
        choices=["default", "mgmoe"],
        help="What type of fusion model to use.",
    )

    parser_.add_argument(
        "--kernel_width",
        type=int,
        default=12,
        help="Base width of the convolutional / locally-connected kernels used.",
    )

    parser_.add_argument(
        "--down_stride",
        type=int,
        default=4,
        help="Down stride to use common over the network when using CNN models.",
    )

    parser_.add_argument(
        "--dilation_factor",
        type=int,
        default=1,
        help="Factor to dilate convolutions by in each successive block when using CNN "
        "models.",
    )

    parser_.add_argument(
        "--first_kernel_expansion",
        type=int,
        default=1,
        help="factor by which to expand the first kernel in the network",
    )

    parser_.add_argument(
        "--first_stride_expansion",
        type=int,
        default=1,
        help="factor by which to expand the first stride in the network",
    )

    parser_.add_argument(
        "--first_channel_expansion",
        type=int,
        default=1,
        help="Factor by which to expand the first stride in the network. "
        "Used for CNN and locally-connected models.",
    )

    parser_.add_argument(
        "--channel_exp_base",
        type=int,
        default=5,
        help="Exponential base for channels in first layer (i.e. default is 2**5). "
        "Used for CNN and locally-connected models.",
    )

    # TODO: Better help message.
    parser_.add_argument(
        "--layers",
        type=int,
        nargs="+",
        help="Number of layers in models where it applies.",
    )

    parser_.add_argument(
        "--rb_do", type=float, default=0.0, help="Dropout in residual blocks."
    )

    parser_.add_argument(
        "--fc_do",
        type=float,
        default=0.0,
        help="Dropout before last fully connected layer.",
    )

    parser_.add_argument(
        "--sa",
        action="store_true",
        help="Whether to add self attention to the network, only applies to CNN.",
    )

    parser_.add_argument(
        "--omics_sources",
        type=str,
        nargs="*",
        help="Which one-hot omics sources to load samples from for training. Can "
        "either be (a) a folder in which files will be gathered from the folder "
        "recursively or (b) a simple text file with each line having a path for "
        "a sample array",
    )

    parser_.add_argument(
        "--omics_names",
        type=str,
        nargs="*",
        help="Names for the omics sources passed in the --omics_sources argument.",
    )

    parser_.add_argument(
        "--valid_size",
        type=float,
        default=0.05,
        help="Size if the validaton set, if float then uses a percentage. If int, "
        "then raw counts.",
    )

    parser_.add_argument(
        "--weighted_sampling_column",
        type=str,
        default=None,
        nargs="*",
        help="Target column to apply weighted sampling on. Only applies to categorical "
        "columns. Passing in 'all' here will use an average of all the target columns.",
    )

    parser_.add_argument(
        "--na_augment_perc",
        default=0.0,
        type=float,
        help="Percentage of array to make missing when using na_augmentation.",
    )

    parser_.add_argument(
        "--na_augment_prob",
        default=0.5,
        type=float,
        help="Probability of applying an na_augmentation of percentage as given in "
        "--na_augment_perc.",
    )

    parser_.add_argument(
        "--mixing_alpha",
        default=0.0,
        type=float,
        help="Alpha parameter used for mixing (higher means more mixing).",
    )

    parser_.add_argument(
        "--mixing_type",
        default=None,
        type=str,
        choices=["mixup", "cutmix-block", "cutmix-uniform"],
        help="Type of mixing to apply when using mixup and similar approaches.",
    )

    parser_.add_argument(
        "--label_file", type=str, required=True, help="Which file to load labels from."
    )

    parser_.add_argument(
        "--label_parsing_chunk_size",
        type=int,
        default=None,
        help="Number of rows to load at a time from label file before processing.",
    )

    parser_.add_argument(
        "--target_con_columns",
        nargs="*",
        default=[],
        help="Continuous target columns in label file.",
    )
    parser_.add_argument(
        "--target_cat_columns",
        nargs="*",
        default=[],
        help="Categorical target columns in label file.",
    )

    parser_.add_argument(
        "--extra_cat_columns",
        type=str,
        nargs="+",
        default=[],
        help="What categorical columns from --label_file to include as input.",
    )

    parser_.add_argument(
        "--extra_con_columns",
        type=str,
        nargs="+",
        default=[],
        help="What continuous columns from --label_file to include as input.",
    )

    parser_.add_argument(
        "--snp_file",
        type=str,
        default="infer",
        help="File to load SNPs from (.bim format).",
    )

    parser_.add_argument(
        "--memory_dataset",
        action="store_true",
        help="Whether to load all sample into memory during training.",
    )

    parser_.add_argument(
        "--sample_interval",
        type=int,
        default=None,
        help="Iteration interval to perform validation.",
    )

    parser_.add_argument(
        "--checkpoint_interval",
        type=int,
        default=None,
        help="Iteration interval to checkpoint model.",
    )

    parser_.add_argument(
        "--n_saved_models",
        type=int,
        default=None,
        help="Number of top N models to saved during training.",
    )

    parser_.add_argument(
        "--run_name",
        required=True,
        type=str,
        help="Name of the current run.",
    )

    parser_.add_argument(
        "--gpu_num", type=str, default="0", help="Which GPU to run (according to CUDA)."
    )

    parser_.add_argument(
        "--multi_gpu",
        action="store_true",
        help="Whether to run the training on multiple GPUs for the current node.",
    )

    parser_.add_argument(
        "--get_acts",
        action="store_true",
        help="Whether to compute activations w.r.t. inputs.",
    )

    parser_.add_argument(
        "--act_classes",
        default=None,
        nargs="+",
        help="Classes to use for activation maps.",
    )

    parser_.add_argument(
        "--max_acts_per_class",
        default=None,
        type=int,
        help="Maximum number of samples per class to gather for activation analysis. "
        "Good to use when modelling on imbalanced data.",
    )

    parser_.add_argument(
        "--act_every_sample_factor",
        type=int,
        default=1,
        help="Controls whether the activations are computed at every sample interval "
        "(=1), every other sample interval (=2), etc. Useful when computing the "
        "activations takes a long time and we don't want to do it every time we "
        "evaluate.",
    )

    parser_.add_argument(
        "--debug",
        action="store_true",
        help="Whether to run in debug mode.",
    )

    parser_.add_argument(
        "--no_pbar",
        action="store_true",
        help="Whether to not use progress bars. Useful when stdout/stderr is written "
        "to files.",
    )

    parser_.add_argument(
        "--plot_skip_steps",
        type=int,
        default=200,
        help="How many iterations to skip in in plots.",
    )
    return parser_


def _get_optimizer_cl_arg_choices():
    """
    Currently just going to hardcode the main default optimizers. Later we can do
    something fancy with inspect and issubclass of Optimizer to get names of all
    PyTorch built-in optimizers.
    """
    base_optimizer_dict = get_base_optimizers_dict()
    default = list(base_optimizer_dict.keys())
    external = _get_custom_opt_names()
    return default + external


def _get_custom_opt_names():
    # import here to keep separated from main codebase
    from torch_optimizer import _NAME_OPTIM_MAP as CUSTOM_OPT_NAME_MAP

    custom_optim_list = list(CUSTOM_OPT_NAME_MAP.keys())
    custom_optim_list = [i for i in custom_optim_list if i != "lookahead"]
    return custom_optim_list


def modify_train_arguments(cl_args: argparse.Namespace) -> argparse.Namespace:

    cl_args_copy = copy(cl_args)

    if cl_args_copy.valid_size > 1.0:
        cl_args_copy.valid_size = int(cl_args_copy.valid_size)

    cl_args_copy.device = (
        "cuda:" + cl_args_copy.gpu_num if torch.cuda.is_available() else "cpu"
    )

    # benchmark breaks if we run it with multiple GPUs
    if not cl_args_copy.multi_gpu:
        torch.backends.cudnn.benchmark = True
    else:
        logger.debug("Setting device to cuda:0 since running with multiple GPUs.")
        cl_args_copy.device = "cuda:0"

    cl_args_copy = append_data_source_prefixes(cl_args=cl_args_copy)

    return cl_args_copy


def append_data_source_prefixes(cl_args: argparse.Namespace) -> argparse.Namespace:
    cl_args_copy = copy(cl_args)

    keys = cl_args.__dict__.keys()
    keys_to_change = [i for i in keys if i.endswith("_names")]

    for key in keys_to_change:
        new_names = []
        prev_names = cl_args_copy.__getattribute__(key)

        if not prev_names:
            continue

        key_type_name = key.split("_names")[0]

        for name in prev_names:
            new_name = key_type_name + "_" + name
            new_names.append(new_name)

        cl_args_copy.__setattr__(key, new_names)

    return cl_args_copy



from dataclasses import dataclass
from typing import TYPE_CHECKING

import torch
from torch import nn

if TYPE_CHECKING:
    from eir.train import DataDimensions


@dataclass
class MLPModelConfig:
    fc_repr_dim: int
    data_dimensions: "DataDimensions"


class MLPModel(nn.Module):
    def __init__(self, model_config: MLPModelConfig):
        super().__init__()

        self.model_config = model_config

        self.fc_0 = nn.Linear(
            self.fc_1_in_features, self.model_config.fc_repr_dim, bias=False
        )

        self._init_weights()

    @property
    def fc_1_in_features(self) -> int:
        return self.model_config.data_dimensions.num_elements()

    @property
    def l1_penalized_weights(self) -> torch.Tensor:
        return self.fc_0.weight

    @property
    def num_out_features(self) -> int:
        return self.model_config.fc_repr_dim

    def _init_weights(self):
        pass

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        out = input.view(input.shape[0], -1)

        out = self.fc_0(out)

        return out



import argparse
import json
from argparse import Namespace
from copy import copy
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from random import sample
from typing import Union, List, Dict, Sequence, Callable, Iterable, Tuple, Type, Any

import dill
import joblib
import numpy as np
import pandas as pd
import torch
from aislib.misc_utils import get_logger, ensure_path_exists
from sklearn.preprocessing import LabelEncoder
from torch import nn
from torch.utils.data import DataLoader

import eir.visualization.visualization_funcs as vf
from eir.configuration import append_data_source_prefixes
from eir.data_load import datasets, label_setup
from eir.data_load.data_utils import get_target_columns_generator
from eir.data_load.datasets import (
    al_datasets,
)
from eir.data_load.label_setup import (
    al_label_dict,
    al_label_transformers_object,
    al_target_columns,
    al_label_transformers,
    al_all_column_ops,
    gather_ids_from_tabular_file,
    transform_label_df,
    TabularFileInfo,
)
from eir.interpretation.interpretation import (
    activation_analysis_wrapper,
    get_no_background_samples_for_shap_objects,
)
from eir.models.model_training_utils import gather_pred_outputs_from_dloader
from eir.models.models_linear import LinearModel, LinearModelConfig
from eir.models.tabular.tabular import al_emb_lookup_dict
from eir.train import (
    get_train_config_serialization_path,
    prepare_base_batch_default,
    Hooks,
    get_tabular_target_label_data,
    get_tabular_inputs_label_data,
    get_fusion_class_from_cl_args,
    get_fusion_kwargs_from_cl_args,
    al_num_outputs_per_target,
    DataDimensions,
)
from eir.train_utils.evaluation import PerformancePlotConfig
from eir.train_utils.metrics import al_metric_record_dict, calculate_batch_metrics
from eir.train_utils.utils import get_run_folder
from eir.train_utils.utils import load_transformers

torch.manual_seed(0)
np.random.seed(0)

logger = get_logger(name=__name__, tqdm_compatible=True)


def main():
    predict_cl_args_ = get_predict_cl_args()
    run_predict(predict_cl_args=predict_cl_args_)


def run_predict(predict_cl_args: Namespace):

    run_folder = Path(predict_cl_args.model_path).parents[1]
    train_config = _load_serialized_train_config(run_folder=run_folder)

    predict_config = get_default_predict_config(
        loaded_train_config=train_config,
        predict_cl_args=predict_cl_args,
    )

    predict(predict_config=predict_config, predict_cl_args=predict_cl_args)

    if predict_cl_args.get_acts:
        _compute_predict_activations(
            train_config=train_config,
            predict_config=predict_config,
            predict_cl_args=predict_cl_args,
        )


def predict(
    predict_config: "PredictConfig",
    predict_cl_args: Namespace,
) -> None:

    all_preds, all_labels, all_ids = gather_pred_outputs_from_dloader(
        data_loader=predict_config.test_dataloader,
        batch_prep_hook=predict_config.hooks.predict_stages.base_prepare_batch,
        batch_prep_hook_kwargs={"predict_config": predict_config},
        model=predict_config.model,
        with_labels=predict_cl_args.evaluate,
    )

    if predict_cl_args.evaluate:
        metrics = calculate_batch_metrics(
            target_columns=predict_config.target_columns,
            outputs=all_preds,
            labels=all_labels,
            mode="val",
            metric_record_dict=predict_config.metrics,
        )
        serialize_prediction_metrics(
            output_folder=Path(predict_cl_args.output_folder), metrics=metrics
        )

    target_columns_gen = get_target_columns_generator(
        target_columns=predict_config.target_columns
    )

    for target_column_type, target_column_name in target_columns_gen:

        target_preds = all_preds[target_column_name]
        predictions = _parse_predictions(
            target_column_type=target_column_type, target_preds=target_preds
        )

        target_labels = None
        if all_labels:
            target_labels = all_labels[target_column_name].cpu().numpy()

        cur_target_transformer = predict_config.target_transformers[target_column_name]
        classes = _get_target_classnames(
            transformer=cur_target_transformer, target_column=target_column_name
        )

        output_folder = Path(predict_cl_args.output_folder, target_column_name)
        ensure_path_exists(path=output_folder, is_folder=True)

        merged_predictions = _merge_ids_predictions_and_labels(
            ids=all_ids,
            predictions=predictions,
            labels=target_labels,
            prediction_classes=classes,
        )

        _save_predictions(
            df_predictions=merged_predictions,
            outfolder=output_folder,
        )

        if predict_cl_args.evaluate:
            cur_labels = all_labels[target_column_name].cpu().numpy()

            plot_config = PerformancePlotConfig(
                val_outputs=predictions,
                val_labels=cur_labels,
                val_ids=all_ids,
                iteration=0,
                column_name=target_column_name,
                column_type=target_column_type,
                target_transformer=cur_target_transformer,
                output_folder=output_folder,
            )

            vf.gen_eval_graphs(plot_config=plot_config)


def _merge_ids_predictions_and_labels(
    ids: Sequence[str],
    predictions: np.ndarray,
    labels: np.ndarray,
    prediction_classes: Union[Sequence[str], None] = None,
    label_column_name: str = "True Label",
) -> pd.DataFrame:
    df = pd.DataFrame()

    df["ID"] = ids
    df = df.set_index("ID")

    df[label_column_name] = labels

    if prediction_classes is None:
        prediction_classes = [f"Score Class {i}" for i in range(predictions.shape[1])]

    df[prediction_classes] = predictions

    return df


def serialize_prediction_metrics(
    output_folder: Path, metrics: Dict[str, Dict[str, np.ndarray]]
):
    with open(str(output_folder / "calculated_metrics.json"), "w") as outfile:
        parsed = _convert_dict_values_to_python_objects(object_=metrics)
        json.dump(obj=parsed, fp=outfile)


def _convert_dict_values_to_python_objects(object_):
    """
    Needed since json does not serialize numpy dtypes natively.
    """

    if isinstance(object_, np.number):
        return object_.item()
    elif isinstance(object_, dict):
        for key, value in object_.items():
            object_[key] = _convert_dict_values_to_python_objects(object_=value)

    return object_


@dataclass
class PredictConfig:
    train_cl_args_overloaded: Namespace
    test_dataset: datasets.DiskDataset
    target_columns: al_target_columns
    target_transformers: al_label_transformers
    test_dataloader: DataLoader
    model: nn.Module
    hooks: "PredictHooks"
    metrics: al_metric_record_dict


@dataclass
class PredictHooks:

    predict_stages: "PredictHookStages"
    custom_column_label_parsing_ops: al_all_column_ops = None


@dataclass
class PredictHookStages:

    al_hook = Callable[..., Dict]
    al_hooks = [Iterable[al_hook]]

    base_prepare_batch: al_hooks
    model_forward: al_hooks


@dataclass
class PredictLabels:
    label_dict: al_label_dict
    transformers: al_label_transformers


def get_default_predict_config(
    loaded_train_config: Namespace,
    predict_cl_args: Namespace,
) -> PredictConfig:

    train_cl_args_overloaded, predict_cl_args = _converge_train_and_predict_cl_args(
        train_cl_args=loaded_train_config.cl_args, predict_cl_args=predict_cl_args
    )

    default_train_hooks = loaded_train_config.hooks

    target_labels, tabular_input_labels = None, None
    if predict_cl_args.evaluate:
        test_ids = gather_ids_from_tabular_file(
            file_path=train_cl_args_overloaded.label_file
        )

        label_ops = default_train_hooks.custom_column_label_parsing_ops
        target_labels, tabular_input_labels = get_target_and_extra_labels_for_predict(
            train_cl_args_overloaded=train_cl_args_overloaded,
            custom_column_label_parsing_ops=label_ops,
            ids=test_ids,
        )

    test_dataset = _set_up_default_test_dataset(
        cl_args=train_cl_args_overloaded,
        data_dimensions=loaded_train_config.data_dimensions,
        target_labels_dict=target_labels.label_dict,
        tabular_inputs_labels_dict=tabular_input_labels.label_dict,
    )

    test_dataloader = DataLoader(
        dataset=test_dataset,
        batch_size=train_cl_args_overloaded.batch_size,
        shuffle=False,
        num_workers=train_cl_args_overloaded.dataloader_workers,
    )

    if train_cl_args_overloaded.model_type == "linear":
        model_class = LinearModel
        model_config = LinearModelConfig(
            input_name=list(loaded_train_config.data_dimensions.keys())[0],
            target_cat_columns=train_cl_args_overloaded.target_cat_columns,
            target_con_columns=train_cl_args_overloaded.target_con_columns,
            data_dimensions=list(loaded_train_config.data_dimensions.values())[0],
        )
        model_init_kwargs = {"model_config": model_config}
        model = _load_model(
            model_path=Path(predict_cl_args.model_path),
            model_class=model_class,
            model_init_kwargs=model_init_kwargs,
            device=train_cl_args_overloaded.device,
        )

    else:
        func = _get_fusion_model_class_and_kwargs_from_cl_args
        fusion_model_class, fusion_model_kwargs = func(
            train_cl_args=train_cl_args_overloaded,
            num_outputs_per_target=loaded_train_config.num_outputs_per_target,
            tabular_input_transformers=tabular_input_labels.transformers,
            omics_data_dimensions=loaded_train_config.data_dimensions,
        )

        model = _load_model(
            model_path=Path(predict_cl_args.model_path),
            model_class=fusion_model_class,
            model_init_kwargs=fusion_model_kwargs,
            device=train_cl_args_overloaded.device,
        )
    assert not model.training

    default_predict_hooks = _get_default_predict_hooks(train_hooks=default_train_hooks)
    test_config = PredictConfig(
        train_cl_args_overloaded=train_cl_args_overloaded,
        test_dataset=test_dataset,
        target_columns=loaded_train_config.target_columns,
        target_transformers=loaded_train_config.target_transformers,
        test_dataloader=test_dataloader,
        model=model,
        hooks=default_predict_hooks,
        metrics=loaded_train_config.metrics,
    )
    return test_config


def _load_serialized_train_config(run_folder: Path) -> Namespace:
    train_config_path = get_train_config_serialization_path(run_folder=run_folder)
    with open(train_config_path, "rb") as infile:
        train_config = dill.load(file=infile)

    return train_config


def get_target_and_extra_labels_for_predict(
    train_cl_args_overloaded: Namespace,
    custom_column_label_parsing_ops: al_all_column_ops,
    ids: Sequence[str],
) -> Tuple[PredictLabels, PredictLabels]:
    """
    NOTE:   This can be extended to more tabular data, including other files, if we
            update the parameters slightly.
    """

    target_info = get_tabular_target_label_data(cl_args=train_cl_args_overloaded)

    target_labels = get_labels_for_predict(
        run_name=train_cl_args_overloaded.run_name,
        tabular_file_info=target_info,
        custom_column_label_parsing_ops=custom_column_label_parsing_ops,
        ids=ids,
    )

    tabular_input_info = get_tabular_inputs_label_data(cl_args=train_cl_args_overloaded)

    if tabular_input_info.con_columns or tabular_input_info.cat_columns:
        tabular_input_labels = get_labels_for_predict(
            run_name=train_cl_args_overloaded.run_name,
            tabular_file_info=tabular_input_info,
            custom_column_label_parsing_ops=custom_column_label_parsing_ops,
            ids=ids,
        )
    else:
        tabular_input_labels = PredictLabels(label_dict={}, transformers={})

    return target_labels, tabular_input_labels


def get_labels_for_predict(
    run_name: str,
    tabular_file_info: TabularFileInfo,
    custom_column_label_parsing_ops: al_all_column_ops,
    ids: Sequence[str],
) -> PredictLabels:

    all_columns = list(tabular_file_info.cat_columns) + list(
        tabular_file_info.con_columns
    )

    if not all_columns:
        raise ValueError(f"No columns specified in {tabular_file_info}.")

    transformers = load_transformers(
        run_name=run_name,
        transformers_to_load=all_columns,
    )

    df_labels = _load_labels_for_predict(
        tabular_info=tabular_file_info,
        ids_to_keep=ids,
        custom_label_ops=custom_column_label_parsing_ops,
    )
    labels_dict = parse_labels_for_testing(
        tabular_info=tabular_file_info,
        df_labels_test=df_labels,
        label_transformers=transformers,
    )

    labels = PredictLabels(label_dict=labels_dict, transformers=transformers)

    return labels


def _get_default_predict_hooks(train_hooks: "Hooks") -> PredictHooks:
    stages = PredictHookStages(
        base_prepare_batch=[_hook_default_predict_prepare_batch],
        model_forward=[train_hooks.step_func_hooks.model_forward],
    )
    predict_hooks = PredictHooks(
        predict_stages=stages,
        custom_column_label_parsing_ops=train_hooks.custom_column_label_parsing_ops,
    )

    return predict_hooks


def _hook_default_predict_prepare_batch(
    predict_config: "PredictConfig",
    loader_batch,
    *args,
    **kwargs,
):
    batch = prepare_base_batch_default(
        loader_batch=loader_batch,
        cl_args=predict_config.train_cl_args_overloaded,
        target_columns=predict_config.target_columns,
        model=predict_config.model,
    )

    state_updates = {"batch": batch}

    return state_updates


def _parse_predictions(
    target_column_type: str, target_preds: torch.Tensor
) -> np.ndarray:
    if target_column_type == "cat":
        predictions = target_preds.cpu().numpy()
    else:
        predictions = target_preds.cpu().numpy()

    return predictions


def _get_target_classnames(
    transformer: al_label_transformers_object, target_column: str
):
    if isinstance(transformer, LabelEncoder):
        return transformer.classes_
    return [target_column]


def _load_cl_args_config(cl_args_config_path: Path) -> Namespace:
    with open(str(cl_args_config_path), "r") as infile:
        loaded_cl_args = json.load(infile)

    return Namespace(**loaded_cl_args)


def _get_fusion_model_class_and_kwargs_from_cl_args(
    train_cl_args: Namespace,
    num_outputs_per_target: al_num_outputs_per_target,
    tabular_input_transformers: al_label_transformers,
    omics_data_dimensions: Dict[str, DataDimensions],
) -> Tuple[Type[nn.Module], Dict[str, Any]]:

    fusion_model_class = get_fusion_class_from_cl_args(
        fusion_model_type=train_cl_args.fusion_model_type
    )

    fusion_model_kwargs = get_fusion_kwargs_from_cl_args(
        cl_args=train_cl_args,
        omics_data_dimensions=omics_data_dimensions,
        num_outputs_per_target=num_outputs_per_target,
        tabular_label_transformers=tabular_input_transformers,
    )

    return fusion_model_class, fusion_model_kwargs


def _load_model(
    model_path: Path,
    model_class: Type[nn.Module],
    model_init_kwargs: Dict,
    device: str,
) -> torch.nn.Module:

    model = model_class(**model_init_kwargs)

    model = _load_model_weights(
        model=model, model_state_dict_path=model_path, device=device
    )

    model.eval()

    return model


def _load_model_weights(
    model: nn.Module, model_state_dict_path: Path, device: str
) -> nn.Module:
    device_for_load = torch.device(device)
    model.load_state_dict(
        state_dict=torch.load(model_state_dict_path, map_location=device_for_load)
    )
    model = model.to(device=device_for_load)

    return model


def _load_saved_embeddings_dict(
    embed_columns: Union[None, List[str]], run_name: str
) -> Union[None, al_emb_lookup_dict]:
    embeddings_dict = None

    run_folder = get_run_folder(run_name)

    if embed_columns:
        model_embeddings_path = run_folder / "extra_inputs" / "embeddings.save"
        embeddings_dict = joblib.load(model_embeddings_path)

    return embeddings_dict


def _converge_train_and_predict_cl_args(
    train_cl_args: Namespace, predict_cl_args: Namespace
) -> Tuple[Namespace, Namespace]:

    train_cl_args_copy = copy(train_cl_args)

    train_cl_args_overloaded = _overload_train_cl_args_for_predict(
        train_cl_args=train_cl_args_copy, predict_cl_args=predict_cl_args
    )
    train_cl_args_overloaded_prefix_renamed = append_data_source_prefixes(
        cl_args=train_cl_args_overloaded
    )

    train_keys = train_cl_args_overloaded_prefix_renamed.__dict__.keys()
    predict_cl_args_filtered = _remove_keys_from_namespace(
        namespace=predict_cl_args, keys_to_remove=train_keys
    )

    return train_cl_args_overloaded_prefix_renamed, predict_cl_args_filtered


def _overload_train_cl_args_for_predict(
    train_cl_args: Namespace, predict_cl_args: Namespace
) -> Namespace:

    train_overloaded = copy(train_cl_args)
    train_keys = set(train_overloaded.__dict__.keys())

    predict_arg_iter = predict_cl_args.__dict__.items()
    for predict_argument_name, predict_argument_value in predict_arg_iter:

        if predict_argument_name in train_keys:
            train_overloaded.__setattr__(predict_argument_name, predict_argument_value)

    return train_overloaded


def _remove_keys_from_namespace(
    namespace: Namespace, keys_to_remove: Iterable[str]
) -> Namespace:
    new_namespace_kwargs = {}

    keys_to_remove_set = set(keys_to_remove)

    for key_name, key_value in namespace.__dict__.items():
        if key_name not in keys_to_remove_set:
            new_namespace_kwargs[key_name] = key_value

    filtered_namespace = Namespace(**new_namespace_kwargs)

    return filtered_namespace


def _load_labels_for_predict(
    tabular_info: TabularFileInfo,
    ids_to_keep: Sequence[str],
    custom_label_ops: al_all_column_ops = None,
) -> pd.DataFrame:

    parse_wrapper = label_setup.get_label_parsing_wrapper(
        label_parsing_chunk_size=tabular_info.parsing_chunk_size
    )
    df_labels_test = parse_wrapper(
        label_file_tabular_info=tabular_info,
        ids_to_keep=ids_to_keep,
        custom_label_ops=custom_label_ops,
    )

    return df_labels_test


def parse_labels_for_testing(
    tabular_info: TabularFileInfo,
    df_labels_test: pd.DataFrame,
    label_transformers: al_label_transformers,
) -> al_label_dict:

    con_transformers = {
        k: v for k, v in label_transformers.items() if k in tabular_info.con_columns
    }
    train_con_column_means = _prep_missing_con_dict(con_transformers=con_transformers)

    df_labels_test = label_setup.handle_missing_label_values_in_df(
        df=df_labels_test,
        cat_label_columns=tabular_info.cat_columns,
        con_label_columns=tabular_info.con_columns,
        con_manual_values=train_con_column_means,
        name="test_df",
    )

    df_labels_test_transformed = transform_label_df(
        df_labels=df_labels_test, label_transformers=label_transformers
    )

    test_labels_dict = df_labels_test_transformed.to_dict("index")

    return test_labels_dict


def _prep_missing_con_dict(con_transformers: al_label_transformers) -> Dict[str, float]:

    train_means = {
        column: transformer.mean_[0] for column, transformer in con_transformers.items()
    }

    return train_means


def _set_up_default_test_dataset(
    cl_args: Namespace,
    data_dimensions: Dict[str, "DataDimensions"],
    target_labels_dict: Union[None, al_label_dict],
    tabular_inputs_labels_dict: Union[None, al_label_dict],
) -> al_datasets:
    """
    :param cl_args: Training CL arguments with the data_source
    replaced with testing one.
    :param target_labels_dict: None if we are predicting on unknown data,
    otherwise a dictionary of labels (if evaluating on test set).
    :return: Dataset instance to be used for loading test samples.
    """

    test_dataset_kwargs = datasets.construct_default_dataset_kwargs_from_cl_args(
        cl_args=cl_args,
        target_labels_dict=target_labels_dict,
        data_dimensions=data_dimensions,
        tabular_labels_dict=tabular_inputs_labels_dict,
        na_augment=False,
    )

    test_dataset = datasets.DiskDataset(**test_dataset_kwargs)

    return test_dataset


def _save_predictions(df_predictions: pd.DataFrame, outfolder: Path) -> None:
    df_predictions.to_csv(outfolder / "predictions.csv")


def get_predict_cl_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model_path",
        type=str,
        required=True,
        help="Path to model to use for predictions.",
    )

    parser.add_argument(
        "--batch_size", type=int, default=64, help="size of the batches"
    )

    parser.add_argument(
        "--label_file", type=str, help="Which file to load labels from."
    )

    parser.add_argument(
        "--omics_sources",
        type=str,
        nargs="*",
        help="Which one-hot omics sources to load samples from for predicting. Can "
        "either be (a) a folder in which files will be gathered from the folder "
        "recursively or (b) a simple text file with each line having a path for "
        "a sample array",
    )

    parser.add_argument(
        "--omics_names",
        type=str,
        nargs="*",
        help="Names for the omics sources passed in the --omics_sources argument.",
    )

    parser.add_argument("--evaluate", dest="evaluate", action="store_true")
    parser.set_defaults(evaluate=False)

    parser.add_argument(
        "--output_folder", type=str, help="Where to save prediction results."
    )

    parser.add_argument(
        "--device",
        type=str,
        default="cpu",
        choices=["gpu", "cpu"],
        help="Which device (CPU or GPU) to run" "the inference on.",
    )

    parser.add_argument(
        "--gpu_num",
        type=str,
        default="0",
        help="Which GPU to run (according to CUDA) if running with flat "
        "'--device gpu'.",
    )

    parser.add_argument(
        "--dataloader_workers",
        type=int,
        default=0,
        help="Number of workers for dataloader.",
    )

    parser.add_argument(
        "--get_acts", action="store_true", help="Whether to generate activation maps."
    )

    parser.add_argument(
        "--act_classes",
        default=None,
        nargs="+",
        help="Classes to use for activation maps.",
    )

    parser.add_argument(
        "--max_acts_per_class",
        default=None,
        type=int,
        help="Maximum number of samples per class to gather for activation analysis. "
        "Good to use when modelling on imbalanced data.",
    )

    cl_args = parser.parse_args()

    device = _parse_predict_device_cl_argument(cl_args=cl_args)
    cl_args.device = device

    return cl_args


def _parse_predict_device_cl_argument(cl_args: argparse.Namespace) -> str:
    if cl_args.device == "gpu":
        device = "cuda:" + cl_args.gpu_num if torch.cuda.is_available() else "cpu"
        if cl_args.device == "cpu":
            logger.warning(
                "Device CL input was 'gpu' with gpu_num '%s', "
                "but number of available GPUs is %d. Reverting to CPU for "
                "now.",
                cl_args.gpu_num,
                torch.cuda.device_count(),
            )
    else:
        device = "cpu"

    return device


def _compute_predict_activations(
    train_config: Namespace, predict_config: PredictConfig, predict_cl_args: Namespace
) -> None:

    overloaded_train_config = _overload_train_config_for_predict_activations(
        train_config=train_config,
        predict_config=predict_config,
        predict_cl_args=predict_cl_args,
    )

    background_dataloader = _get_background_dataloader(
        batch_size=predict_cl_args.batch_size,
        dataloader_workers=predict_cl_args.dataloader_workers,
        cl_args=train_config.cl_args,
        data_dimensions=train_config.data_dimensions,
        label_parsing_ops=train_config.hooks.custom_column_label_parsing_ops,
    )

    activation_outfolder_callable = partial(
        _get_predict_activation_outfolder_target,
        predict_outfolder=Path(predict_cl_args.output_folder),
    )

    activation_analysis_wrapper(
        model=predict_config.model,
        train_config=overloaded_train_config,
        outfolder_target_callable=activation_outfolder_callable,
        dataset_to_interpret=predict_config.test_dataset,
        background_loader=background_dataloader,
    )


def _overload_train_config_for_predict_activations(
    train_config: Namespace, predict_config: PredictConfig, predict_cl_args: Namespace
) -> Namespace:
    train_config_copy = copy(train_config)

    train_config_copy.model = predict_config.model

    train_config_copy.cl_args, _ = _converge_train_and_predict_cl_args(
        train_cl_args=train_config.cl_args, predict_cl_args=predict_cl_args
    )

    train_config_copy.cl_args.act_classes = predict_cl_args.act_classes
    train_config_copy.cl_args.max_acts_per_class = predict_cl_args.max_acts_per_class

    return train_config_copy


def _get_background_dataloader(
    batch_size: int,
    dataloader_workers: int,
    cl_args: Namespace,
    data_dimensions,
    label_parsing_ops,
):
    background_size = get_no_background_samples_for_shap_objects(batch_size=batch_size)

    train_ids = gather_ids_from_tabular_file(file_path=cl_args.label_file)
    train_ids_sampled = sample(population=train_ids, k=background_size)

    target_labels, tabular_input_labels = get_target_and_extra_labels_for_predict(
        train_cl_args_overloaded=cl_args,
        custom_column_label_parsing_ops=label_parsing_ops,
        ids=train_ids_sampled,
    )

    background_dataset = _set_up_default_test_dataset(
        cl_args=cl_args,
        data_dimensions=data_dimensions,
        target_labels_dict=target_labels.label_dict,
        tabular_inputs_labels_dict=tabular_input_labels.label_dict,
    )

    background_loader = DataLoader(
        dataset=background_dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=dataloader_workers,
    )

    return background_loader


def _get_predict_activation_outfolder_target(
    predict_outfolder: Path, column_name: str, input_name: str
) -> Path:
    activation_outfolder = predict_outfolder / column_name / "activations" / input_name
    ensure_path_exists(path=activation_outfolder, is_folder=True)

    return activation_outfolder


if __name__ == "__main__":
    main()



from copy import copy
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Union, Dict, Callable, Sequence, Tuple

import numpy as np
import torch
from torch import nn

from eir.data_load.data_utils import Batch
from eir.data_load.label_setup import al_target_columns

if TYPE_CHECKING:
    from train import al_training_labels_target, al_criterions, Config

al_target_values = Union[torch.LongTensor, torch.Tensor]
al_int_tensors = Union[
    torch.ByteTensor,
    torch.CharTensor,
    torch.ShortTensor,
    torch.IntTensor,
    torch.LongTensor,
    torch.BoolTensor,
]


@dataclass
class MixingObject:
    targets: "al_training_labels_target"
    targets_permuted: "al_training_labels_target"
    lambda_: float
    permuted_indexes: al_int_tensors


def get_mix_data_hook(mixing_type: str):
    mixing_func_mapping = _get_mixing_function_map()
    mixing_func = mixing_func_mapping.get(mixing_type)

    bound_hook = partial(hook_default_mix_data, mixing_func=mixing_func)

    return bound_hook


def hook_default_mix_data(
    config: "Config", state: Dict, mixing_func: Callable, *args, **kwargs
) -> Dict:

    batch = state["batch"]

    mixing_info = get_mixing_info(
        mixing_alpha=config.cl_args.mixing_alpha,
        batch_size=config.cl_args.batch_size,
        target_labels=batch.target_labels,
        target_columns=config.target_columns,
    )

    mixed_inputs = {}

    for input_name, input_data in batch.inputs.items():
        if input_name.startswith("omics_"):

            mixed_omics = mixup_omics_data(
                inputs=input_data,
                mixing_func=mixing_func,
                mixing_info=mixing_info,
            )
            mixed_inputs[input_name] = mixed_omics

        elif input_name.startswith("tabular_"):

            mixed_tabular_input_tensor = mixup_tensor(
                tensor=input_data,
                lambda_=mixing_info.lambda_,
                random_batch_indices_to_mix=mixing_info.permuted_indexes,
            )
            mixed_inputs[input_name] = mixed_tabular_input_tensor

    batch_mixed = Batch(
        inputs=mixed_inputs,
        target_labels=batch.target_labels,
        ids=batch.ids,
    )

    state_updates = {
        "batch": batch_mixed,
        "mixing_info": mixing_info,
    }

    return state_updates


def get_mixing_info(
    mixing_alpha: float,
    batch_size: int,
    target_labels: "al_training_labels_target",
    target_columns: al_target_columns,
) -> MixingObject:
    lambda_ = _sample_lambda(mixing_alpha=mixing_alpha)

    permuted_indexes = get_random_batch_indices_to_mix(batch_size=batch_size)
    targets_permuted = mixup_all_targets(
        targets=target_labels,
        random_index_for_mixing=permuted_indexes,
        target_columns=target_columns,
    )

    mixing_info = MixingObject(
        targets=target_labels,
        targets_permuted=targets_permuted,
        lambda_=lambda_,
        permuted_indexes=permuted_indexes,
    )

    return mixing_info


def _sample_lambda(mixing_alpha: float) -> float:
    if mixing_alpha > 0:
        lambda_ = np.random.beta(mixing_alpha, mixing_alpha)
    else:
        lambda_ = 1.0

    return lambda_


def hook_mix_loss(config: "Config", state: Dict, *args, **kwargs) -> Dict:

    mixed_losses = calc_all_mixed_losses(
        target_columns=config.target_columns,
        criterions=config.criterions,
        outputs=state["model_outputs"],
        mixed_object=state["mixing_info"],
    )

    state_updates = {"per_target_train_losses": mixed_losses}

    return state_updates


def _get_mixing_function_map():
    mapping = {
        "cutmix-uniform": uniform_cutmix_omics_input,
        "cutmix-block": block_cutmix_omics_input,
        "mixup": mixup_input,
    }
    return mapping


def mixup_omics_data(
    inputs: torch.Tensor,
    mixing_func: Callable[[torch.Tensor, float, torch.Tensor], torch.Tensor],
    mixing_info: MixingObject,
) -> torch.Tensor:
    """
    NOTE: **This function will modify the inputs in-place**

    The original inputs (inputs) will be lost when calling this function, unless
    they have been explicitly copied and stored in another variable before this call.

    This is because we do not want to clone the input tensor in this (or any functions
    called within this) function as it might mean a large memory overhead if the inputs
    are large.

    An exception is when we use the "vanilla" MixUp, as that calculates a new tensor
    instead of cut-pasting inside an already existing tensor.
    """
    assert inputs.dim() == 4, "Should be called with 4 dimensions."

    mixed_inputs = mixing_func(
        input_batch=inputs,
        lambda_=mixing_info.lambda_,
        random_batch_indices_to_mix=mixing_info.permuted_indexes,
    )

    return mixed_inputs


def get_random_batch_indices_to_mix(batch_size: int) -> al_int_tensors:
    return torch.randperm(batch_size).to(dtype=torch.long)


def mixup_input(
    input_batch: torch.Tensor,
    lambda_: float,
    random_batch_indices_to_mix: al_int_tensors,
) -> torch.Tensor:
    """
    This function is to delegate arguments from mixup_snp_data to a general mixup
    function that is does not necessarily have an 'input_batch' argument.
    """
    mixed_input = mixup_tensor(
        tensor=input_batch,
        lambda_=lambda_,
        random_batch_indices_to_mix=random_batch_indices_to_mix,
    )

    return mixed_input


def mixup_tensor(
    tensor: torch.Tensor,
    lambda_: float,
    random_batch_indices_to_mix: Sequence[al_int_tensors],
) -> torch.Tensor:

    mixed_tensor = (
        lambda_ * tensor + (1.0 - lambda_) * tensor[random_batch_indices_to_mix, :]
    )
    return mixed_tensor


def block_cutmix_omics_input(
    input_batch: torch.Tensor, lambda_: float, random_batch_indices_to_mix: torch.Tensor
) -> torch.Tensor:

    cut_start, cut_end = get_block_cutmix_indices(
        input_length=input_batch.shape[-1], lambda_=lambda_
    )
    target_to_cut = input_batch[random_batch_indices_to_mix, :]
    cut_part = target_to_cut[..., cut_start:cut_end]

    # Caution: input_ will be modified as well since no .clone() below here
    cutmixed_x = input_batch
    cutmixed_x[..., cut_start:cut_end] = cut_part

    return cutmixed_x


def get_block_cutmix_indices(input_length: int, lambda_: float) -> Tuple[int, int]:
    mixin_coefficient = 1.0 - lambda_
    num_snps_to_mix = int(round(input_length * mixin_coefficient))
    random_index_start = np.random.choice(max(1, input_length - num_snps_to_mix))
    random_index_end = random_index_start + num_snps_to_mix
    return random_index_start, random_index_end


def uniform_cutmix_omics_input(
    input_batch: torch.Tensor,
    lambda_: float,
    random_batch_indices_to_mix: torch.Tensor,
) -> torch.Tensor:

    target_to_mix = input_batch[random_batch_indices_to_mix, :]

    random_snp_indices_to_mix = get_uniform_cutmix_indices(
        input_length=input_batch.shape[-1], lambda_=lambda_
    )
    cut_part = target_to_mix[..., random_snp_indices_to_mix]

    # Caution: input_ will be modified as well since no .clone() below here
    cutmixed_x = input_batch
    cutmixed_x[..., random_snp_indices_to_mix] = cut_part

    return cutmixed_x


def get_uniform_cutmix_indices(input_length: int, lambda_: float) -> torch.Tensor:
    mixin_coefficient = 1.0 - lambda_
    num_snps_to_mix = int(round(input_length * mixin_coefficient))
    random_to_mix = np.random.choice(input_length, num_snps_to_mix, replace=False)
    random_to_mix = torch.tensor(random_to_mix, dtype=torch.long)

    return random_to_mix


def mixup_all_targets(
    targets: "al_training_labels_target",
    random_index_for_mixing: al_int_tensors,
    target_columns: al_target_columns,
) -> "al_training_labels_target":
    targets_permuted = copy(targets)

    all_target_cols = target_columns["cat"] + target_columns["con"]
    for target_col in all_target_cols:

        cur_targets = targets_permuted[target_col]
        cur_targets_permuted = mixup_targets(
            targets=cur_targets, random_index_for_mixing=random_index_for_mixing
        )
        targets_permuted[target_col] = cur_targets_permuted

    return targets_permuted


def mixup_targets(
    targets: al_target_values,
    random_index_for_mixing: torch.Tensor,
) -> al_target_values:

    targets_permuted = copy(targets)
    targets_permuted = targets_permuted[random_index_for_mixing]

    return targets_permuted


def calc_all_mixed_losses(
    target_columns: al_target_columns,
    criterions: "al_criterions",
    outputs: Dict[str, torch.Tensor],
    mixed_object: MixingObject,
):

    losses = {}

    all_target_columns = target_columns["con"] + target_columns["cat"]
    for target_col in all_target_columns:
        cur_loss = calc_mixed_loss(
            criterion=criterions[target_col],
            outputs=outputs[target_col],
            targets=mixed_object.targets[target_col],
            targets_permuted=mixed_object.targets_permuted[target_col],
            lambda_=mixed_object.lambda_,
        )
        losses[target_col] = cur_loss

    return losses


def calc_mixed_loss(
    criterion: Union[nn.CrossEntropyLoss, nn.MSELoss],
    outputs: torch.Tensor,
    targets: torch.Tensor,
    targets_permuted: torch.Tensor,
    lambda_: float,
) -> torch.Tensor:

    base_loss = lambda_ * criterion(input=outputs, target=targets)
    permuted_loss = (1.0 - lambda_) * criterion(input=outputs, target=targets_permuted)

    total_loss = base_loss + permuted_loss

    return total_loss


def make_random_omics_columns_missing(
    omics_array: torch.Tensor, percentage: float = 0.05, probability: float = 1.0
) -> torch.Tensor:
    random_draw = np.random.uniform()
    if random_draw > probability:
        return omics_array

    n_snps = omics_array.shape[2]
    n_to_drop = (int(n_snps * percentage),)
    random_to_drop = np.random.choice(n_snps, n_to_drop, replace=False)
    random_to_drop = torch.tensor(random_to_drop, dtype=torch.long)

    missing_arr = torch.tensor([False, False, False, True], dtype=torch.bool).reshape(
        -1, 1
    )
    omics_array[:, :, random_to_drop] = missing_arr

    return omics_array



import argparse
from argparse import Namespace
from dataclasses import _DataclassParams
from typing import Union, Type, Dict, Any, Protocol, TYPE_CHECKING

from eir.models.omics.models_cnn import CNNModel, CNNModelConfig
from eir.models.omics.models_mlp import MLPModel, MLPModelConfig
from eir.models.omics.models_split_mlp import (
    SplitMLPModel,
    FullySplitMLPModel,
    SplitMLPModelConfig,
    FullySplitMLPModelConfig,
)

if TYPE_CHECKING:
    from eir.train import DataDimensions


al_models_classes = Union[
    Type["CNNModel"],
    Type["MLPModel"],
    Type["SplitMLPModel"],
    Type["FullySplitMLPModel"],
]

al_models = Union[
    "CNNModel",
    "MLPModel",
    "SplitMLPModel",
    "FullySplitMLPModel",
]


def _get_model_mapping() -> Dict[str, al_models_classes]:
    mapping = {
        "cnn": CNNModel,
        "mlp": MLPModel,
        "mlp-split": SplitMLPModel,
        "genome-local-net": FullySplitMLPModel,
    }

    return mapping


def get_model_class(model_type: str) -> al_models_classes:
    mapping = _get_model_mapping()
    return mapping[model_type]


class Dataclass(Protocol):
    __dataclass_fields__: Dict
    __dataclass_params__: _DataclassParams


def _get_dataclass_mapping() -> Dict[str, Type[Dataclass]]:
    mapping = {
        "cnn": CNNModelConfig,
        "mlp": MLPModelConfig,
        "mlp-split": SplitMLPModelConfig,
        "genome-local-net": FullySplitMLPModelConfig,
    }

    return mapping


def get_model_config_dataclass(model_type: str) -> Type[Dataclass]:
    mapping = _get_dataclass_mapping()
    return mapping[model_type]


def get_omics_model_init_kwargs(
    model_type: str, cl_args: Namespace, data_dimensions: "DataDimensions"
) -> Dict[str, Any]:
    """
    See: https://github.com/python/mypy/issues/5374 for type hint issue.

    Possibly split / extend this function later to account for other kwargs that just
    model_config, to allow for more flexibility in model instantiation (not restricting
    to just model_config object).
    """

    kwargs = {}

    model_config_dataclass = get_model_config_dataclass(model_type=model_type)
    model_config_dataclass_kwargs = match_namespace_to_dataclass(
        namespace=cl_args, data_class=model_config_dataclass
    )

    if "data_dimensions" in model_config_dataclass.__dataclass_fields__.keys():
        model_config_dataclass_kwargs["data_dimensions"] = data_dimensions

    dataclass_instance = model_config_dataclass(**model_config_dataclass_kwargs)

    kwargs["model_config"] = dataclass_instance

    return kwargs


def match_namespace_to_dataclass(
    namespace: argparse.Namespace, data_class: Type[Dataclass]
) -> Dict[str, Any]:
    dataclass_kwargs = {}
    field_names = data_class.__dataclass_fields__.keys()

    for field_name in field_names:
        if hasattr(namespace, field_name):
            dataclass_kwargs[field_name] = getattr(namespace, field_name)

    return dataclass_kwargs



from pathlib import Path
from typing import Dict, Sequence, TYPE_CHECKING, List
from collections import defaultdict
from textwrap import wrap
import warnings

import torch
import shap
import matplotlib.pyplot as plt
from matplotlib import MatplotlibDeprecationWarning
import numpy as np
import pandas as pd
import seaborn as sns
from aislib.misc_utils import get_logger, ensure_path_exists

from eir.interpretation.interpret_omics import _get_target_class_name
from eir.train_utils.utils import load_transformers

if TYPE_CHECKING:
    from eir.train import Config
    from eir.data_load.label_setup import al_label_transformers_object
    from eir.interpretation.interpretation import SampleActivation

logger = get_logger(__name__)


def analyze_tabular_input_activations(
    config: "Config",
    input_name: str,
    target_column_name: str,
    target_column_type: str,
    activation_outfolder: Path,
    all_activations: Sequence["SampleActivation"],
):

    tabular_model = config.model.modules_to_fuse[input_name]
    activation_tensor_slices = set_up_tabular_tensor_slices(
        cat_input_columns=config.cl_args.extra_cat_columns,
        con_input_columns=config.cl_args.extra_con_columns,
        embedding_module=tabular_model,
    )

    parsed_activations = parse_tabular_activations_for_feature_importance(
        activation_tensor_slices=activation_tensor_slices,
        all_activations=all_activations,
        input_name=input_name,
    )
    df_activations = get_tabular_activation_df(parsed_activations=parsed_activations)
    plot_tabular_activations(
        df_activations=df_activations, activation_outfolder=activation_outfolder
    )

    all_activations_class_stratified = stratify_activations_by_target_classes(
        all_activations=all_activations,
        target_transformer=config.target_transformers[target_column_name],
        target_column=target_column_name,
        column_type=target_column_type,
    )
    for class_name, class_activations in all_activations_class_stratified.items():

        cur_class_outfolder = activation_outfolder / class_name
        ensure_path_exists(path=cur_class_outfolder, is_folder=True)

        if config.cl_args.extra_con_columns:

            cat_to_con_cutoff = get_cat_to_con_cutoff_from_slices(
                slices=activation_tensor_slices,
                cat_input_columns=config.cl_args.extra_cat_columns,
            )
            continuous_shap = _gather_continuous_shap_values(
                all_activations=class_activations,
                cat_to_con_cutoff=cat_to_con_cutoff,
                input_name=input_name,
            )
            continuous_inputs = _gather_continuous_inputs(
                all_activations=class_activations,
                cat_to_con_cutoff=cat_to_con_cutoff,
                con_names=config.cl_args.extra_con_columns,
                input_name=input_name,
            )
            plot_tabular_beeswarm(
                shap_values=continuous_shap,
                features=continuous_inputs,
                activation_outfolder=cur_class_outfolder,
                class_name=class_name,
            )

        for cat_column in config.cl_args.extra_cat_columns:

            categorical_shap = _gather_categorical_shap_values(
                all_activations=class_activations,
                cur_slice=activation_tensor_slices[cat_column],
                input_name=input_name,
            )
            categorical_inputs = _gather_categorical_inputs(
                all_activations=class_activations,
                cat_name=cat_column,
                input_name=input_name,
            )
            cat_column_transformers = load_transformers(
                run_name=config.cl_args.run_name, transformers_to_load=[cat_column]
            )

            categorical_inputs_mapped = map_categorical_labels_to_names(
                cat_column_transformers=cat_column_transformers,
                cat_column=cat_column,
                categorical_inputs=categorical_inputs,
            )

            plot_tabular_categorical_feature(
                shap_values=categorical_shap,
                features=categorical_inputs_mapped,
                feature_name_to_plot=cat_column,
                class_name=class_name,
                activation_outfolder=cur_class_outfolder,
            )


def set_up_tabular_tensor_slices(
    cat_input_columns: Sequence[str],
    con_input_columns: Sequence[str],
    embedding_module: torch.nn.Module,
) -> Dict[str, slice]:
    """
    TODO: Probably better to pass in Dict[column_name, nn.Embedding]
    """
    slices = {}

    current_index = 0
    for cat_column in cat_input_columns:
        cur_embedding = getattr(embedding_module, "embed_" + cat_column)
        cur_embedding_dim = cur_embedding.embedding_dim

        slice_start = current_index
        slice_end = current_index + cur_embedding_dim
        slices[cat_column] = slice(slice_start, slice_end)

        current_index = slice_end

    for con_column in con_input_columns:
        slice_start = current_index
        slice_end = slice_start + 1

        slices[con_column] = slice(slice_start, slice_end)
        current_index = slice_end

    return slices


def stratify_activations_by_target_classes(
    all_activations: Sequence["SampleActivation"],
    target_transformer: "al_label_transformers_object",
    target_column: str,
    column_type: str,
) -> Dict[str, Sequence["SampleActivation"]]:
    all_activations_class_stratified = defaultdict(list)

    for sample in all_activations:
        cur_label_name = _get_target_class_name(
            sample_label=sample.sample_info.target_labels[target_column],
            target_transformer=target_transformer,
            column_type=column_type,
            target_column_name=target_column,
        )
        all_activations_class_stratified[cur_label_name].append(sample)

    return all_activations_class_stratified


def get_cat_to_con_cutoff_from_slices(
    slices: Dict[str, slice], cat_input_columns: Sequence[str]
) -> int:
    cutoff = 0

    for cat_column in cat_input_columns:
        cur_cat_slice = slices[cat_column]
        slice_size = cur_cat_slice.stop - cur_cat_slice.start
        cutoff += slice_size

    return cutoff


def parse_tabular_activations_for_feature_importance(
    activation_tensor_slices: Dict,
    all_activations: Sequence["SampleActivation"],
    input_name: str,
) -> Dict[str, List[float]]:
    """
    Note we need to use abs here to get the absolute feature importance, before
    we sum so different signs don't cancel each other out.
    """
    column_activations = {column: [] for column in activation_tensor_slices.keys()}

    for sample_activation in all_activations:
        sample_acts = sample_activation.sample_activations[input_name].squeeze()

        for column in activation_tensor_slices.keys():
            cur_slice = activation_tensor_slices[column]
            cur_column_activations = np.abs(sample_acts[cur_slice]).sum()
            column_activations[column].append(cur_column_activations)

    finalized_activations = {}

    for column, aggregated_activations in column_activations.items():
        mean_activation = np.array(column_activations[column]).mean()
        finalized_activations[column] = [mean_activation]

    return finalized_activations


def get_tabular_activation_df(
    parsed_activations: Dict[str, List[float]]
) -> pd.DataFrame:
    df_activations = pd.DataFrame.from_dict(
        parsed_activations, orient="index", columns=["Shap_Value"]
    )
    df_activations = df_activations.sort_values(by="Shap_Value", ascending=False)

    return df_activations


def plot_tabular_activations(
    df_activations: pd.DataFrame, activation_outfolder: Path
) -> None:

    sns_plot = sns.barplot(
        x=df_activations["Shap_Value"],
        y=df_activations.index,
        palette="Blues_d",
    )
    plt.tight_layout()
    sns_figure = sns_plot.get_figure()
    sns_figure.set_size_inches(10, 0.5 * len(df_activations))
    sns_figure.savefig(
        str(activation_outfolder / "feature_importance.png"), bbox_inches="tight"
    )

    plt.close("all")


def _gather_continuous_inputs(
    all_activations: Sequence["SampleActivation"],
    cat_to_con_cutoff: int,
    con_names: Sequence[str],
    input_name: str,
) -> pd.DataFrame:
    con_inputs = []

    for sample in all_activations:
        cur_full_input = sample.sample_info.inputs[input_name]
        cur_con_input_part = cur_full_input.squeeze()[cat_to_con_cutoff:]
        con_inputs.append(cur_con_input_part)

    con_inputs_array = np.array([np.array(i.cpu()) for i in con_inputs])
    df = pd.DataFrame(con_inputs_array, columns=con_names)

    return df


def _gather_continuous_shap_values(
    all_activations: Sequence["SampleActivation"],
    cat_to_con_cutoff: int,
    input_name: str,
) -> np.ndarray:
    """
    We need to expand_dims in the case where we only have one input value, otherwise
    shap will not accept a vector.
    """
    con_acts = []

    for sample in all_activations:
        cur_full_input = sample.sample_activations[input_name]
        cur_con_input_part = cur_full_input.squeeze()[cat_to_con_cutoff:]
        con_acts.append(cur_con_input_part)

    array_squeezed = np.array(con_acts).squeeze()

    if len(array_squeezed.shape) == 1:
        return np.expand_dims(array_squeezed, 1)

    return array_squeezed


def _gather_categorical_shap_values(
    all_activations: Sequence["SampleActivation"],
    cur_slice: slice,
    input_name: str,
) -> np.ndarray:
    """
    We need to expand_dims in the case where we only have one input value, otherwise
    shap will not accept a vector.
    """
    cat_acts = []

    for sample in all_activations:
        cur_full_input = sample.sample_activations[input_name]
        cur_cat_input_part = cur_full_input.squeeze()[cur_slice]
        cur_cat_summed_act = np.sum(np.array(cur_cat_input_part))
        cat_acts.append(cur_cat_summed_act)

    array_squeezed = np.array(cat_acts).squeeze()

    if len(array_squeezed.shape) == 1:
        return np.expand_dims(array_squeezed, 1)

    return array_squeezed


def _gather_categorical_inputs(
    all_activations: Sequence["SampleActivation"],
    cat_name: str,
    input_name: str,
) -> pd.DataFrame:
    cat_inputs = []

    for sample in all_activations:

        cur_raw_cat_input = sample.raw_tabular_inputs[input_name][cat_name]
        cur_cat_input_part = cur_raw_cat_input.squeeze()
        cat_inputs.append(cur_cat_input_part)

    cat_inputs_array = np.array([np.array(i.cpu()) for i in cat_inputs])
    df = pd.DataFrame(cat_inputs_array, columns=[cat_name])

    return df


def map_categorical_labels_to_names(
    cat_column_transformers: Dict, cat_column: str, categorical_inputs: pd.DataFrame
) -> pd.DataFrame:
    cat_column_transformer = cat_column_transformers[cat_column]

    categorical_inputs_copy = categorical_inputs.copy()

    categorical_inputs_copy[cat_column] = cat_column_transformer.inverse_transform(
        categorical_inputs_copy[cat_column]
    )

    return categorical_inputs_copy


def plot_tabular_categorical_feature(
    shap_values: np.ndarray,
    features: pd.DataFrame,
    feature_name_to_plot: str,
    class_name: str,
    activation_outfolder: Path,
):
    """
    We catch the warnings there because shap is causing them in the dependence plot,
    to avoid filling the screen with warnings.
    """

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=MatplotlibDeprecationWarning)
        warnings.simplefilter("ignore", category=UserWarning)

        _ = shap.dependence_plot(
            ind=feature_name_to_plot,
            shap_values=shap_values,
            features=features,
            interaction_index=None,
            color="black",
        )

    fig = plt.gcf()

    plt.title(feature_name_to_plot, fontsize=14)
    plt.yticks(fontsize=10, wrap=True)

    plt.tight_layout()

    fig_name = f"cat_features_{feature_name_to_plot}_{class_name}.png"
    output_path = str(activation_outfolder / fig_name)

    try:
        fig.savefig(output_path, bbox_inches="tight")

    # See: https://github.com/matplotlib/matplotlib/issues/17579
    except ZeroDivisionError:
        logger.error(
            "Encountered ZeroDivisionError when saving %s. Skipping.", output_path
        )

    plt.close("all")


def plot_tabular_beeswarm(
    shap_values: np.ndarray,
    features: pd.DataFrame,
    activation_outfolder: Path,
    class_name: str,
):

    _ = shap.summary_plot(
        shap_values=shap_values,
        features=features,
        show=False,
        title=class_name,
        max_display=20,
    )

    fig = plt.gcf()

    ax = _get_shap_main_axis(figure=fig)
    ax_wrapped_ytick_labels = _get_wrapped_ax_ytick_text(axis=ax)
    ax.set_yticklabels(ax_wrapped_ytick_labels)

    plt.title(class_name, fontsize=14)
    plt.yticks(fontsize=10, wrap=True)

    plt.tight_layout()
    fig.savefig(
        str(activation_outfolder / f"con_features_beeswarm_{class_name}.png"),
        bbox_inches="tight",
    )

    plt.close("all")


def _get_shap_main_axis(figure: plt.Figure):
    return figure.axes[0]


def _get_wrapped_ax_ytick_text(axis: plt.Axes):
    labels = [item.get_text() for item in axis.get_yticklabels()]
    labels_wrapped = ["\n".join(wrap(label, 20)) for label in labels]

    return labels_wrapped



import logging
from functools import wraps
from pathlib import Path
from typing import (
    List,
    Dict,
    TYPE_CHECKING,
    Sequence,
    Callable,
    Iterable,
    Union,
    Any,
    Tuple,
)

import joblib
from aislib.misc_utils import get_logger, ensure_path_exists
from ignite.engine import Engine

from eir.data_load import label_setup
from eir.data_load.label_setup import al_label_transformers

logger = get_logger(name=__name__, tqdm_compatible=True)

if TYPE_CHECKING:
    from eir.data_load.label_setup import al_label_dict


def get_extra_labels_from_ids(
    labels_dict: "al_label_dict", cur_ids: List[str], target_columns: List[str]
) -> List[Dict[str, str]]:
    """
    Returns a batch in same order as cur_ids.
    """
    extra_labels = []
    for sample_id in cur_ids:
        cur_labels_all = labels_dict.get(sample_id)
        cur_labels_extra = {
            k: v for k, v in cur_labels_all.items() if k in target_columns
        }
        extra_labels.append(cur_labels_extra)

    return extra_labels


def get_run_folder(run_name: str) -> Path:
    return Path("runs", run_name)


def prep_sample_outfolder(run_name: str, column_name: str, iteration: int) -> Path:
    sample_outfolder = (
        get_run_folder(run_name) / "results" / column_name / "samples" / str(iteration)
    )
    ensure_path_exists(sample_outfolder, is_folder=True)

    return sample_outfolder


def configure_root_logger(run_name: str):

    logfile_path = get_run_folder(run_name=run_name) / "logging_history.log"

    ensure_path_exists(logfile_path)
    file_handler = logging.FileHandler(str(logfile_path))
    file_handler.setLevel(logging.DEBUG)

    formatter = logging.Formatter(
        "%(asctime)s - %(levelname)s - %(name)s - %(message)s", "%H:%M:%S"
    )
    file_handler.setFormatter(formatter)

    logging.getLogger("").addHandler(file_handler)


def validate_handler_dependencies(handler_dependencies: Sequence[Callable]):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            argument_iterable = tuple(args) + tuple(kwargs.values())

            engine_objects = [i for i in argument_iterable if isinstance(i, Engine)]
            assert len(engine_objects) == 1

            engine_object = engine_objects[0]

            for dep in handler_dependencies:
                if not engine_object.has_event_handler(dep):
                    raise MissingHandlerDependencyError(
                        f"Dependency '{dep.__name__}' missing from engine."
                    )

            func_output = func(*args, **kwargs)
            return func_output

        return wrapper

    return decorator


class MissingHandlerDependencyError(Exception):
    pass


def call_hooks_stage_iterable(
    hook_iterable: Iterable[Callable],
    common_kwargs: Dict,
    state: Union[None, Dict[str, Any]],
):
    for hook in hook_iterable:
        _, state = state_registered_hook_call(
            hook_func=hook, **common_kwargs, state=state
        )

    return state


def state_registered_hook_call(
    hook_func: Callable,
    state: Union[Dict[str, Any], None],
    *args,
    **kwargs,
) -> Tuple[Any, Dict[str, Any]]:

    if state is None:
        state = {}

    state_updates = hook_func(state=state, *args, **kwargs)

    state = {**state, **state_updates}

    return state_updates, state


def load_transformers(
    run_name: str, transformers_to_load: Union[Sequence[str], None]
) -> al_label_transformers:

    run_folder = get_run_folder(run_name=run_name)
    all_transformers = (i.stem for i in (run_folder / "transformers").iterdir())

    iterable = transformers_to_load if transformers_to_load else all_transformers

    label_transformers = {}
    for transformer_name in iterable:
        target_transformer_path = label_setup.get_transformer_path(
            run_path=run_folder, transformer_name=transformer_name
        )
        target_transformer_object = joblib.load(filename=target_transformer_path)
        label_transformers[transformer_name] = target_transformer_object

    return label_transformers



from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict, TYPE_CHECKING

import numpy as np
import pandas as pd
import torch
from aislib.misc_utils import get_logger
from ignite.engine import Engine
from scipy.special import softmax
from sklearn.preprocessing import LabelEncoder, StandardScaler

from eir.data_load.data_utils import get_target_columns_generator
from eir.data_load.label_setup import al_label_transformers_object
from eir.models import model_training_utils
from eir.train_utils import metrics
from eir.train_utils import utils
from eir.visualization import visualization_funcs as vf

if TYPE_CHECKING:
    from eir.train_utils.train_handlers import HandlerConfig
    from eir.train import Config

logger = get_logger(name=__name__, tqdm_compatible=True)


def validation_handler(engine: Engine, handler_config: "HandlerConfig") -> None:
    """
    A bit hacky how we manually attach metrics here, but that's because we
    don't want to evaluate as a running average (i.e. do it in the step
    function), but rather run over the whole validation dataset as we do
    in this function.

    TODO: Streamline with hooks.
    """
    c = handler_config.config
    cl_args = c.cl_args
    iteration = engine.state.iteration

    c.model.eval()
    gather_preds = model_training_utils.gather_pred_outputs_from_dloader

    val_outputs_total, val_target_labels, val_ids_total = gather_preds(
        data_loader=c.valid_loader,
        batch_prep_hook=c.hooks.step_func_hooks.base_prepare_batch,
        batch_prep_hook_kwargs={"config": c},
        model=c.model,
        with_labels=True,
    )
    c.model.train()

    val_target_labels = model_training_utils.parse_target_labels(
        target_columns=c.target_columns, device=cl_args.device, labels=val_target_labels
    )

    val_losses = c.loss_function(inputs=val_outputs_total, targets=val_target_labels)
    val_loss_avg = metrics.aggregate_losses(losses_dict=val_losses)

    eval_metrics_dict = metrics.calculate_batch_metrics(
        target_columns=c.target_columns,
        outputs=val_outputs_total,
        labels=val_target_labels,
        mode="val",
        metric_record_dict=c.metrics,
    )

    eval_metrics_dict_w_loss = metrics.add_loss_to_metrics(
        target_columns=c.target_columns,
        losses=val_losses,
        metric_dict=eval_metrics_dict,
    )

    eval_metrics_dict_w_avgs = metrics.add_multi_task_average_metrics(
        batch_metrics_dict=eval_metrics_dict_w_loss,
        target_columns=c.target_columns,
        loss=val_loss_avg.item(),
        performance_average_functions=c.metrics["averaging_functions"],
    )

    write_eval_header = True if iteration == cl_args.sample_interval else False
    metrics.persist_metrics(
        handler_config=handler_config,
        metrics_dict=eval_metrics_dict_w_avgs,
        iteration=iteration,
        write_header=write_eval_header,
        prefixes={"metrics": "validation_", "writer": "validation"},
    )

    save_evaluation_results_wrapper(
        val_outputs=val_outputs_total,
        val_labels=val_target_labels,
        val_ids=val_ids_total,
        iteration=iteration,
        config=handler_config.config,
    )


def save_evaluation_results_wrapper(
    val_outputs: Dict[str, torch.Tensor],
    val_labels: Dict[str, torch.Tensor],
    val_ids: List[str],
    iteration: int,
    config: "Config",
):

    target_columns_gen = get_target_columns_generator(config.target_columns)
    transformers = config.target_transformers

    for column_type, column_name in target_columns_gen:
        cur_sample_outfolder = utils.prep_sample_outfolder(
            run_name=config.cl_args.run_name,
            column_name=column_name,
            iteration=iteration,
        )

        cur_val_outputs = val_outputs[column_name].cpu().numpy()
        cur_val_labels = val_labels[column_name].cpu().numpy()

        plot_config = PerformancePlotConfig(
            val_outputs=cur_val_outputs,
            val_labels=cur_val_labels,
            val_ids=val_ids,
            iteration=iteration,
            column_name=column_name,
            column_type=column_type,
            target_transformer=transformers[column_name],
            output_folder=cur_sample_outfolder,
        )

        save_evaluation_results(plot_config=plot_config)


@dataclass
class PerformancePlotConfig:
    val_outputs: np.ndarray
    val_labels: np.ndarray
    val_ids: List[str]
    iteration: int
    column_name: str
    column_type: str
    target_transformer: al_label_transformers_object
    output_folder: Path


def save_evaluation_results(
    plot_config: PerformancePlotConfig,
) -> None:

    pc = plot_config

    common_args = {
        "val_outputs": pc.val_outputs,
        "val_labels": pc.val_labels,
        "val_ids": pc.val_ids,
        "outfolder": pc.output_folder,
        "transformer": pc.target_transformer,
    }

    vf.gen_eval_graphs(plot_config=pc)

    if pc.column_type == "cat":
        get_most_wrong_wrapper(**common_args)
    elif pc.column_type == "con":
        scale_and_save_regression_preds(**common_args)


def get_most_wrong_wrapper(
    val_labels: np.ndarray,
    val_outputs: np.ndarray,
    val_ids: List[str],
    transformer: LabelEncoder,
    outfolder: Path,
) -> None:
    val_preds_total = val_outputs.argmax(axis=1)

    some_are_incorrect = (val_labels != val_preds_total).sum() > 0
    if some_are_incorrect:
        df_most_wrong = get_most_wrong_cls_preds(
            val_true=val_labels,
            val_preds=val_preds_total,
            val_outputs=val_outputs,
            ids=np.array(val_ids),
        )

        df_most_wrong = _inverse_numerical_labels_hook(
            df=df_most_wrong, target_transformer=transformer
        )
        _check_df_most_wrong(df=df_most_wrong, outfolder=outfolder)
        df_most_wrong.to_csv(outfolder / "wrong_preds.csv")


def _check_df_most_wrong(df: pd.DataFrame, outfolder: Path) -> None:
    try:
        assert_1_string = "True label equal predicted labels."
        assert not (df["True_Label"] == df["Wrong_Label"]).any(), assert_1_string

        assert (df["True_Prob"] < 0.5).all(), "True predicted over 0.5."

    except AssertionError as e:
        logger.error(
            "Got AssertionError ('%s') when checking for probabilities of wrong "
            "predictions. Something might be weird, or a rare event where probabilities"
            "are exactly equal happened. The file is wrong_preds.csv in %s.",
            e,
            outfolder,
        )


def get_most_wrong_cls_preds(
    val_true: np.ndarray,
    val_preds: np.ndarray,
    val_outputs: np.ndarray,
    ids: np.ndarray,
) -> pd.DataFrame:
    wrong_mask = val_preds != val_true
    wrong_indices = np.where(wrong_mask)[0]
    all_wrong_probs = softmax(val_outputs[wrong_indices], axis=1)

    correct_labels_for_misclassified = val_true[wrong_indices]

    correct_label_prob = all_wrong_probs[
        np.arange(wrong_indices.shape[0]), correct_labels_for_misclassified
    ]
    assert correct_label_prob.max() <= 0.5

    wrong_pred_labels = val_preds[wrong_indices]
    wrong_label_pred_prob = all_wrong_probs[
        np.arange(wrong_indices.shape[0]), wrong_pred_labels
    ]

    columns = ["Sample_ID", "True_Label", "True_Prob", "Wrong_Label", "Wrong_Prob"]
    column_values = [
        ids[wrong_indices],
        correct_labels_for_misclassified,
        correct_label_prob,
        wrong_pred_labels,
        wrong_label_pred_prob,
    ]

    df = pd.DataFrame(columns=columns)

    for col_name, data in zip(columns, column_values):
        df[col_name] = data

    df = df.sort_values(by=["True_Prob"])
    return df


def _inverse_numerical_labels_hook(
    df: pd.DataFrame, target_transformer: LabelEncoder
) -> pd.DataFrame:
    for column in ["True_Label", "Wrong_Label"]:
        df[column] = target_transformer.inverse_transform(df[column])

    return df


def scale_and_save_regression_preds(
    val_labels: np.ndarray,
    val_outputs: np.ndarray,
    val_ids: List[str],
    transformer: StandardScaler,
    outfolder: Path,
) -> None:
    val_labels = transformer.inverse_transform(val_labels).squeeze()
    val_outputs = transformer.inverse_transform(val_outputs).squeeze()

    data = np.array([val_ids, val_labels, val_outputs]).T
    df = pd.DataFrame(data=data, columns=["ID", "Actual", "Predicted"])

    df.to_csv(outfolder / "regression_predictions.csv", index=["ID"])



import argparse
import sys
from dataclasses import dataclass, fields
from functools import partial
from pathlib import Path
from types import SimpleNamespace
from typing import (
    Union,
    Tuple,
    List,
    Dict,
    overload,
    TYPE_CHECKING,
    Callable,
    Iterable,
    Sequence,
    Any,
    Set,
    Type,
)

import dill
import numpy as np
import torch
from aislib.misc_utils import ensure_path_exists
from aislib.misc_utils import get_logger
from ignite.engine import Engine
from sklearn.preprocessing import StandardScaler
from torch import nn
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader, WeightedRandomSampler
from torch.utils.tensorboard import SummaryWriter

from eir.configuration import get_default_cl_args
from eir.data_load import data_utils
from eir.data_load import datasets
from eir.data_load.data_augmentation import hook_mix_loss, get_mix_data_hook
from eir.data_load.data_loading_funcs import get_weighted_random_sampler
from eir.data_load.data_utils import Batch
from eir.data_load.label_setup import (
    al_target_columns,
    al_label_transformers,
    al_all_column_ops,
    get_array_path_iterator,
    set_up_train_and_valid_tabular_data,
    gather_ids_from_tabular_file,
    split_ids,
    TabularFileInfo,
    save_transformer_set,
    Labels,
)
from eir.models import fusion, fusion_mgmoe
from eir.models import model_training_utils
from eir.models.model_training_utils import run_lr_find
from eir.models.models_linear import LinearModel, LinearModelConfig
from eir.models.omics.omics_models import (
    get_model_class,
    get_omics_model_init_kwargs,
    match_namespace_to_dataclass,
)
from eir.models.tabular.tabular import (
    get_tabular_inputs,
    TabularModel,
    get_unique_values_from_transformers,
)
from eir.train_utils import utils
from eir.train_utils.metrics import (
    calculate_batch_metrics,
    calculate_prediction_losses,
    aggregate_losses,
    add_multi_task_average_metrics,
    get_average_history_filepath,
    get_default_metrics,
    hook_add_l1_loss,
    get_uncertainty_loss_hook,
    add_loss_to_metrics,
)
from eir.train_utils.optimizers import (
    get_optimizer,
    get_optimizer_backward_kwargs,
)
from eir.train_utils.train_handlers import HandlerConfig
from eir.train_utils.train_handlers import configure_trainer
from eir.train_utils.utils import (
    call_hooks_stage_iterable,
)

if TYPE_CHECKING:
    from eir.train_utils.metrics import (
        al_step_metric_dict,
        al_metric_record_dict,
    )

# aliases
al_criterions = Dict[str, Union[nn.CrossEntropyLoss, nn.MSELoss]]
# these are all after being collated by torch dataloaders
al_training_labels_target = Dict[str, Union[torch.LongTensor, torch.Tensor]]
al_training_labels_extra = Dict[str, Union[List[str], torch.Tensor]]
al_training_labels_batch = Dict[
    str, Union[al_training_labels_target, al_training_labels_extra]
]
al_dataloader_getitem_batch = Tuple[
    Union[Dict[str, torch.Tensor], Dict[str, Any]],
    al_training_labels_target,
    List[str],
]
al_num_outputs_per_target = Dict[str, int]

torch.manual_seed(0)
np.random.seed(0)

logger = get_logger(name=__name__, tqdm_compatible=True)


def main():
    default_cl_args = get_default_cl_args()
    utils.configure_root_logger(run_name=default_cl_args.run_name)

    default_hooks = get_default_hooks(cl_args_=default_cl_args)
    default_config = get_default_config(cl_args=default_cl_args, hooks=default_hooks)

    run_experiment(cl_args=default_cl_args, config=default_config)


def run_experiment(cl_args: argparse.Namespace, config: "Config") -> None:

    _log_model(model=config.model, l1_weight=cl_args.l1)

    run_folder = utils.get_run_folder(run_name=cl_args.run_name)
    keys_to_serialize = get_default_config_keys_to_serialize()
    serialize_config(
        config=config, run_folder=run_folder, keys_to_serialize=keys_to_serialize
    )

    train(config=config)


@dataclass(frozen=True)
class Config:
    """
    The idea of this class is to keep track of objects that need to be used
    in multiple contexts in different parts of the code (e.g. the train
    dataloader is used to load samples during training, but also as background
    for SHAP activation calculations).
    """

    cl_args: argparse.Namespace
    data_dimensions: Dict[str, "DataDimensions"]
    train_loader: torch.utils.data.DataLoader
    valid_loader: torch.utils.data.DataLoader
    valid_dataset: torch.utils.data.Dataset
    labels_dict: Dict
    target_transformers: al_label_transformers
    target_columns: al_target_columns
    num_outputs_per_target: al_num_outputs_per_target
    model: Union[fusion.FusionModel, nn.DataParallel]
    optimizer: Optimizer
    criterions: al_criterions
    loss_function: Callable
    writer: SummaryWriter
    metrics: "al_metric_record_dict"
    hooks: Union["Hooks", None]


def serialize_config(
    config: "Config", run_folder: Path, keys_to_serialize: Union[Iterable[str], None]
) -> None:
    serialization_path = get_train_config_serialization_path(run_folder=run_folder)
    ensure_path_exists(path=serialization_path)

    filtered_config = filter_config_by_keys(config=config, keys=keys_to_serialize)
    serialize_namespace(namespace=filtered_config, output_path=serialization_path)


def get_train_config_serialization_path(run_folder: Path) -> Path:
    train_config_path = run_folder / "serializations" / "filtered_config.dill"

    return train_config_path


def get_default_config_keys_to_serialize() -> Iterable[str]:
    return (
        "cl_args",
        "data_dimensions",
        "num_outputs_per_target",
        "target_transformers",
        "target_columns",
        "metrics",
        "hooks",
    )


def filter_config_by_keys(
    config: "Config", keys: Union[None, Iterable[str]] = None
) -> SimpleNamespace:
    filtered = {}

    config_fields = (f.name for f in fields(config))
    iterable = keys if keys is not None else config_fields

    for k in iterable:
        filtered[k] = getattr(config, k)

    namespace = SimpleNamespace(**filtered)

    return namespace


def serialize_namespace(namespace: SimpleNamespace, output_path: Path) -> None:
    with open(output_path, "wb") as outfile:
        dill.dump(namespace, outfile)


def get_default_config(
    cl_args: argparse.Namespace, hooks: Union["Hooks", None] = None
) -> "Config":
    run_folder = _prepare_run_folder(run_name=cl_args.run_name)

    target_labels, tabular_input_labels = get_target_and_tabular_input_labels(
        cl_args=cl_args,
        custom_label_parsing_operations=hooks.custom_column_label_parsing_ops,
    )
    save_transformer_set(
        transformers=target_labels.label_transformers, run_folder=run_folder
    )
    save_transformer_set(
        transformers=tabular_input_labels.label_transformers, run_folder=run_folder
    )

    data_dimensions = _gather_all_omics_data_dimensions(
        omics_sources=cl_args.omics_sources, omics_names=cl_args.omics_names
    )

    train_dataset, valid_dataset = datasets.set_up_datasets_from_cl_args(
        cl_args=cl_args,
        data_dimensions=data_dimensions,
        target_labels=target_labels,
        tabular_inputs_labels=tabular_input_labels,
    )

    batch_size = _modify_bs_for_multi_gpu(
        multi_gpu=cl_args.multi_gpu, batch_size=cl_args.batch_size
    )

    train_sampler = get_train_sampler(
        columns_to_sample=cl_args.weighted_sampling_column, train_dataset=train_dataset
    )

    train_dloader, valid_dloader = get_dataloaders(
        train_dataset=train_dataset,
        train_sampler=train_sampler,
        valid_dataset=valid_dataset,
        batch_size=batch_size,
        num_workers=cl_args.dataloader_workers,
    )

    num_outputs_per_target = set_up_num_outputs_per_target(
        target_transformers=target_labels.label_transformers
    )

    model = get_model_from_cl_args(
        cl_args=cl_args,
        omics_data_dimensions=data_dimensions,
        num_outputs_per_target=num_outputs_per_target,
        tabular_label_transformers=tabular_input_labels.label_transformers,
    )

    criterions = _get_criterions(
        target_columns=train_dataset.target_columns, model_type=cl_args.model_type
    )

    writer = get_summary_writer(run_folder=run_folder)

    loss_func = _get_loss_callable(
        criterions=criterions,
    )

    optimizer = get_optimizer(model=model, loss_callable=loss_func, cl_args=cl_args)

    metrics = get_default_metrics(target_transformers=target_labels.label_transformers)

    config = Config(
        cl_args=cl_args,
        data_dimensions=data_dimensions,
        train_loader=train_dloader,
        valid_loader=valid_dloader,
        valid_dataset=valid_dataset,
        labels_dict=train_dataset.target_labels_dict,
        target_transformers=target_labels.label_transformers,
        num_outputs_per_target=num_outputs_per_target,
        target_columns=train_dataset.target_columns,
        model=model,
        optimizer=optimizer,
        criterions=criterions,
        loss_function=loss_func,
        writer=writer,
        metrics=metrics,
        hooks=hooks,
    )

    return config


def get_target_and_tabular_input_labels(
    cl_args: argparse.Namespace, custom_label_parsing_operations: al_all_column_ops
) -> Tuple[Labels, Labels]:
    """
    TODO: Set up support for multiple tabular files.
    """
    all_array_ids = gather_ids_from_tabular_file(file_path=cl_args.label_file)
    train_ids, valid_ids = split_ids(ids=all_array_ids, valid_size=cl_args.valid_size)

    logger.info("Setting up target labels.")
    target_labels_info = get_tabular_target_label_data(cl_args=cl_args)
    target_labels = set_up_train_and_valid_tabular_data(
        tabular_info=target_labels_info,
        custom_label_ops=custom_label_parsing_operations,
        train_ids=train_ids,
        valid_ids=valid_ids,
    )

    tabular_inputs_info = get_tabular_inputs_label_data(cl_args=cl_args)

    n_cat_tabular = len(tabular_inputs_info.cat_columns)
    n_con_tabular = len(tabular_inputs_info.con_columns)
    if n_cat_tabular + n_con_tabular > 0:
        logger.info("Setting up tabular inputs.")
        tabular_inputs = set_up_train_and_valid_tabular_data(
            tabular_info=tabular_inputs_info,
            custom_label_ops=custom_label_parsing_operations,
            train_ids=train_ids,
            valid_ids=valid_ids,
        )
    else:
        tabular_inputs = Labels(train_labels={}, valid_labels={}, label_transformers={})

    return target_labels, tabular_inputs


def get_tabular_target_label_data(cl_args: argparse.Namespace) -> TabularFileInfo:

    table_info = TabularFileInfo(
        file_path=cl_args.label_file,
        con_columns=cl_args.target_con_columns,
        cat_columns=cl_args.target_cat_columns,
        parsing_chunk_size=cl_args.label_parsing_chunk_size,
    )

    return table_info


def get_tabular_inputs_label_data(cl_args: argparse.Namespace) -> TabularFileInfo:

    table_info = TabularFileInfo(
        file_path=cl_args.label_file,
        con_columns=cl_args.extra_con_columns,
        cat_columns=cl_args.extra_cat_columns,
        parsing_chunk_size=cl_args.label_parsing_chunk_size,
    )

    return table_info


def _gather_all_omics_data_dimensions(
    omics_sources: Sequence[str], omics_names: Sequence[str]
) -> Dict[str, "DataDimensions"]:

    data_dimensions = {}

    if not omics_sources and not omics_names:
        return data_dimensions

    assert len(omics_sources) == len(omics_names)

    for source, name in zip(omics_sources, omics_names):
        cur_dimension = _get_data_dimension_from_data_source(data_source=Path(source))
        data_dimensions[name] = cur_dimension

    return data_dimensions


@dataclass
class DataDimensions:
    channels: int
    height: int
    width: int

    def num_elements(self):
        return self.channels * self.height * self.width


def _get_data_dimension_from_data_source(
    data_source: Path,
) -> DataDimensions:
    """
    TODO: Make more dynamic / robust. Also weird to say "width" for a 1D vector.
    """

    iterator = get_array_path_iterator(data_source=data_source)
    path = next(iterator)
    shape = np.load(file=path).shape

    if len(shape) == 1:
        channels, height, width = 1, 1, shape[0]
    elif len(shape) == 2:
        channels, height, width = 1, shape[0], shape[1]
    elif len(shape) == 3:
        channels, height, width = shape
    else:
        raise ValueError("Currently max 3 dimensional inputs supported")

    return DataDimensions(channels=channels, height=height, width=width)


def set_up_num_outputs_per_target(
    target_transformers: al_label_transformers,
) -> al_num_outputs_per_target:

    num_outputs_per_target_dict = {}
    for target_column, transformer in target_transformers.items():
        if isinstance(transformer, StandardScaler):
            num_outputs = 1
        else:
            num_outputs = len(transformer.classes_)

            if num_outputs < 2:
                logger.warning(
                    f"Only {num_outputs} unique values found in categorical label "
                    f"column {target_column} (returned by {transformer}). This means "
                    f"that most likely an error will be raised if e.g. using "
                    f"nn.CrossEntropyLoss as it expects an output dimension of >=2."
                )

        num_outputs_per_target_dict[target_column] = num_outputs

    return num_outputs_per_target_dict


def _prepare_run_folder(run_name: str) -> Path:
    run_folder = utils.get_run_folder(run_name=run_name)
    history_file = get_average_history_filepath(
        run_folder=run_folder, train_or_val_target_prefix="train_"
    )
    if history_file.exists():
        raise FileExistsError(
            f"There already exists a run with that name: {history_file}. Please choose "
            f"a different run name or delete the folder."
        )

    ensure_path_exists(path=run_folder, is_folder=True)

    return run_folder


def _modify_bs_for_multi_gpu(multi_gpu: bool, batch_size: int) -> int:
    if multi_gpu:
        batch_size = torch.cuda.device_count() * batch_size
        logger.info(
            "Batch size set to %d to account for %d GPUs.",
            batch_size,
            torch.cuda.device_count(),
        )

    return batch_size


@overload
def get_train_sampler(
    columns_to_sample: None, train_dataset: datasets.DatasetBase
) -> None:
    ...


@overload
def get_train_sampler(
    columns_to_sample: List[str], train_dataset: datasets.DatasetBase
) -> WeightedRandomSampler:
    ...


def get_train_sampler(columns_to_sample, train_dataset):
    """
    TODO:   Refactor, remove dependency on train_dataset and use instead
            Iterable[Samples], and target_columns directly.
    """
    if columns_to_sample is None:
        return None

    loaded_target_columns = (
        train_dataset.target_columns["con"] + train_dataset.target_columns["cat"]
    )

    is_sample_column_loaded = set(columns_to_sample).issubset(
        set(loaded_target_columns)
    )
    is_sample_all_cols = columns_to_sample == ["all"]

    if not is_sample_column_loaded and not is_sample_all_cols:
        raise ValueError(
            "Weighted sampling from non-loaded columns not supported yet "
            f"(could not find {columns_to_sample})."
        )

    if is_sample_all_cols:
        columns_to_sample = train_dataset.target_columns["cat"]

    train_sampler = get_weighted_random_sampler(
        samples=train_dataset.samples, target_columns=columns_to_sample
    )
    return train_sampler


def get_dataloaders(
    train_dataset: datasets.DatasetBase,
    train_sampler: Union[None, WeightedRandomSampler],
    valid_dataset: datasets.DatasetBase,
    batch_size: int,
    num_workers: int = 0,
) -> Tuple:

    train_dloader = DataLoader(
        dataset=train_dataset,
        batch_size=batch_size,
        sampler=train_sampler,
        shuffle=False if train_sampler else True,
        num_workers=num_workers,
        pin_memory=False,
        drop_last=True,
    )

    valid_dloader = DataLoader(
        dataset=valid_dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        pin_memory=False,
        drop_last=True,
    )

    return train_dloader, valid_dloader


class GetAttrDelegatedDataParallel(nn.DataParallel):
    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError:
            return getattr(self.module, name)


def get_model_from_cl_args(
    cl_args: argparse.Namespace,
    omics_data_dimensions: Dict[str, DataDimensions],
    num_outputs_per_target: al_num_outputs_per_target,
    tabular_label_transformers: al_label_transformers,
) -> Union[nn.Module, nn.DataParallel]:
    """
    Note:   If we have a linear model, we just return that because that takes care of
            fusing tabular and genotype data by itself.

    Note:   This function currently assumes 1 omics source, and one tabular source.
            If we wanted to expand it, we would have to pass in nested configurations
            for each data source (i.e. model), instead of just cl_args. Also we would
            need to have nested data_dimensions, linking to models.
    """

    if cl_args.model_type == "linear":
        assert len(omics_data_dimensions) == 1

        linear_model = _get_linear_model(
            target_cat_columns=cl_args.target_cat_columns,
            target_con_columns=cl_args.target_con_columns,
            data_dimensions=omics_data_dimensions,
            device=cl_args.device,
        )
        return linear_model

    fusion_class = get_fusion_class_from_cl_args(
        fusion_model_type=cl_args.fusion_model_type
    )
    fusion_kwargs = get_fusion_kwargs_from_cl_args(
        cl_args=cl_args,
        omics_data_dimensions=omics_data_dimensions,
        num_outputs_per_target=num_outputs_per_target,
        tabular_label_transformers=tabular_label_transformers,
    )
    fusion_model = fusion_class(**fusion_kwargs)
    fusion_model = fusion_model.to(device=cl_args.device)

    if cl_args.multi_gpu:
        fusion_model = GetAttrDelegatedDataParallel(module=fusion_model)

    return fusion_model


def _get_linear_model(
    target_cat_columns: Sequence[str],
    target_con_columns: Sequence[str],
    data_dimensions: Dict[str, DataDimensions],
    device: str,
) -> LinearModel:
    """
    TODO: Update / fix when we make a more general linear model. Currently it only
          supports one omics input and tabular data.
    """

    assert len(data_dimensions) == 1

    single_input_name = list(data_dimensions.keys())[0]
    single_input_data_dimension = list(data_dimensions.values())[0]

    model_config = LinearModelConfig(
        input_name=single_input_name,
        data_dimensions=single_input_data_dimension,
        target_cat_columns=target_cat_columns,
        target_con_columns=target_con_columns,
    )

    model = LinearModel(model_config=model_config)
    model = model.to(device=device)

    return model


def get_modules_to_fuse_from_cl_args(
    cl_args: argparse.Namespace,
    omics_data_dimensions: Dict[str, DataDimensions],
    tabular_label_transformers: al_label_transformers,
):
    models = nn.ModuleDict()

    for name, dimensions in omics_data_dimensions.items():
        cur_omics_model = get_omics_model_from_cl_args(
            cl_args=cl_args, data_dimensions=dimensions
        )

        models[name] = cur_omics_model

    if cl_args.extra_con_columns or cl_args.extra_cat_columns:
        unique_tabular_values = get_unique_values_from_transformers(
            transformers=tabular_label_transformers,
            keys_to_use=cl_args.extra_cat_columns,
        )

        tabular_model = get_tabular_model(
            cat_columns=cl_args.extra_cat_columns,
            con_columns=cl_args.extra_con_columns,
            device=cl_args.device,
            unique_label_values=unique_tabular_values,
        )
        models["tabular_cl_args"] = tabular_model

    return models


def get_tabular_model(
    cat_columns: Sequence[str],
    con_columns: Sequence[str],
    device: str,
    unique_label_values: Dict[str, Set[str]],
) -> TabularModel:
    tabular_model = TabularModel(
        cat_columns=cat_columns,
        con_columns=con_columns,
        unique_label_values=unique_label_values,
        device=device,
    )

    return tabular_model


def get_omics_model_from_cl_args(
    cl_args: argparse.Namespace, data_dimensions: DataDimensions
):

    omics_model_class = get_model_class(model_type=cl_args.model_type)
    model_init_kwargs = get_omics_model_init_kwargs(
        model_type=cl_args.model_type, cl_args=cl_args, data_dimensions=data_dimensions
    )
    omics_model = omics_model_class(**model_init_kwargs)

    if cl_args.model_type == "cnn":
        assert omics_model.data_size_after_conv >= 8

    omics_model = omics_model.to(device=cl_args.device)

    return omics_model


def get_fusion_class_from_cl_args(
    fusion_model_type: str,
) -> Type[nn.Module]:
    if fusion_model_type == "mgmoe":
        return fusion_mgmoe.MGMoEModel
    elif fusion_model_type == "default":
        return fusion.FusionModel
    raise ValueError(f"Unrecognized fusion model type: {fusion_model_type}.")


def get_fusion_kwargs_from_cl_args(
    cl_args: argparse.Namespace,
    omics_data_dimensions: Dict[str, DataDimensions],
    num_outputs_per_target: al_num_outputs_per_target,
    tabular_label_transformers: al_label_transformers,
) -> Dict[str, Any]:

    kwargs = {}
    modules_to_fuse = get_modules_to_fuse_from_cl_args(
        cl_args=cl_args,
        omics_data_dimensions=omics_data_dimensions,
        tabular_label_transformers=tabular_label_transformers,
    )
    kwargs["modules_to_fuse"] = modules_to_fuse

    kwargs["num_outputs_per_target"] = num_outputs_per_target

    model_dataclass_config_class = fusion.FusionModelConfig
    if cl_args.fusion_model_type == "mgmoe":
        model_dataclass_config_class = fusion_mgmoe.MGMoEModelConfig
    model_dataclass_config_kwargs = match_namespace_to_dataclass(
        namespace=cl_args, data_class=model_dataclass_config_class
    )
    model_config = model_dataclass_config_class(**model_dataclass_config_kwargs)

    kwargs["model_config"] = model_config

    return kwargs


def _check_linear_model_columns(cl_args: argparse.Namespace) -> None:
    num_label_cols = len(cl_args.target_cat_columns + cl_args.target_con_columns)
    if num_label_cols != 1:
        raise NotImplementedError(
            "Linear model only supports one target column currently."
        )

    num_extra_cols = len(cl_args.extra_cat_columns + cl_args.extra_con_columns)
    if num_extra_cols != 0:
        raise NotImplementedError(
            "Extra columns not supported for linear model currently."
        )


def _get_criterions(
    target_columns: al_target_columns, model_type: str
) -> al_criterions:
    criterions_dict = {}

    def get_criterion(column_type_):

        if model_type == "linear":
            if column_type_ == "cat":
                return _calc_bce
            else:
                return partial(_calc_mse, mse_loss_func=nn.MSELoss(reduction="mean"))

        if column_type_ == "con":
            return partial(_calc_mse, mse_loss_func=nn.MSELoss())
        elif column_type_ == "cat":
            return nn.CrossEntropyLoss()

    target_columns_gen = data_utils.get_target_columns_generator(
        target_columns=target_columns
    )

    for column_type, column_name in target_columns_gen:
        criterion = get_criterion(column_type_=column_type)
        criterions_dict[column_name] = criterion

    return criterions_dict


def _calc_bce(input, target):
    # note we use input and not e.g. input_ here because torch uses name "input"
    # in loss functions for compatibility
    bce_loss_func = nn.BCELoss()
    return bce_loss_func(input[:, 1], target.to(dtype=torch.float))


def _calc_mse(input, target, mse_loss_func: nn.MSELoss):
    return mse_loss_func(input=input.squeeze(), target=target.squeeze())


def _get_loss_callable(criterions: al_criterions):

    single_task_loss_func = partial(calculate_prediction_losses, criterions=criterions)
    return single_task_loss_func


def get_summary_writer(run_folder: Path) -> SummaryWriter:
    log_dir = Path(run_folder / "tensorboard_logs")
    writer = SummaryWriter(log_dir=str(log_dir))

    return writer


def _log_model(model: nn.Module, l1_weight: float) -> None:
    """
    TODO: Add summary of parameters
    TODO: Add verbosity option
    """
    no_params = sum(p.numel() for p in model.parameters() if p.requires_grad)

    if l1_weight:
        logger.debug(
            "Penalizing weights of shape %s with L1 loss with weight %f.",
            model.l1_penalized_weights.shape,
            l1_weight,
        )

    logger.info(
        "Starting training with a %s parameter model.", format(no_params, ",.0f")
    )


def train(config: Config) -> None:
    c = config
    cl_args = config.cl_args
    step_hooks = c.hooks.step_func_hooks

    def step(
        engine: Engine,
        loader_batch: Tuple[torch.Tensor, al_training_labels_batch, List[str]],
    ) -> "al_step_metric_dict":
        """
        The output here goes to trainer.output.
        """
        c.model.train()
        c.optimizer.zero_grad()

        base_prepare_inputs_stage = step_hooks.base_prepare_batch
        state = call_hooks_stage_iterable(
            hook_iterable=base_prepare_inputs_stage,
            common_kwargs={"config": config, "loader_batch": loader_batch},
            state=None,
        )
        base_batch = state["batch"]

        post_prepare_inputs_stage = step_hooks.post_prepare_batch
        state = call_hooks_stage_iterable(
            hook_iterable=post_prepare_inputs_stage,
            common_kwargs={"config": config, "loader_batch": base_batch},
            state=state,
        )
        batch = state["batch"]

        model_forward_loss_stage = step_hooks.model_forward
        state = call_hooks_stage_iterable(
            hook_iterable=model_forward_loss_stage,
            common_kwargs={"config": config, "batch": batch},
            state=state,
        )

        loss_stage = step_hooks.loss
        state = call_hooks_stage_iterable(
            hook_iterable=loss_stage,
            common_kwargs={"config": config, "batch": batch},
            state=state,
        )

        optimizer_backward_stage = step_hooks.optimizer_backward
        state = call_hooks_stage_iterable(
            hook_iterable=optimizer_backward_stage,
            common_kwargs={"config": config, "batch": batch},
            state=state,
        )

        metrics_stage = step_hooks.metrics
        state = call_hooks_stage_iterable(
            hook_iterable=metrics_stage,
            common_kwargs={"config": config, "batch": batch},
            state=state,
        )

        return state["metrics"]

    trainer = Engine(process_function=step)

    if cl_args.find_lr:
        logger.info("Running LR find and exiting.")
        run_lr_find(
            trainer_engine=trainer,
            train_dataloader=c.train_loader,
            model=c.model,
            optimizer=c.optimizer,
            output_folder=utils.get_run_folder(run_name=cl_args.run_name),
        )
        sys.exit(0)

    trainer = configure_trainer(trainer=trainer, config=config)

    trainer.run(data=c.train_loader, max_epochs=cl_args.n_epochs)


def get_default_hooks(cl_args_: argparse.Namespace):
    step_func_hooks = _get_default_step_function_hooks(cl_args=cl_args_)
    hooks_object = Hooks(step_func_hooks=step_func_hooks)

    return hooks_object


@dataclass
class Hooks:
    al_handler_attachers = Iterable[Callable[[Engine, HandlerConfig], Engine]]

    step_func_hooks: "StepFunctionHookStages"
    custom_column_label_parsing_ops: al_all_column_ops = None
    custom_handler_attachers: Union[None, al_handler_attachers] = None


def _get_default_step_function_hooks(cl_args: argparse.Namespace):
    """
    TODO: Add validation, inspect that outputs have correct names.
    TODO: Refactor, split into smaller functions e.g. for L1, mixing and uncertainty.
    """

    init_kwargs = _get_default_step_function_hooks_init_kwargs(cl_args=cl_args)

    step_func_hooks = StepFunctionHookStages(**init_kwargs)

    return step_func_hooks


def _get_default_step_function_hooks_init_kwargs(
    cl_args: argparse.Namespace,
) -> Dict[str, Sequence[Callable]]:

    init_kwargs = {
        "base_prepare_batch": [hook_default_prepare_batch],
        "post_prepare_batch": [],
        "model_forward": [hook_default_model_forward],
        "loss": [hook_default_per_target_loss],
        "optimizer_backward": [hook_default_optimizer_backward],
        "metrics": [hook_default_compute_metrics],
    }

    if cl_args.mixing_type is not None:
        logger.debug(
            "Setting up hooks for mixing with %s with α=%.2g.",
            cl_args.mixing_type,
            cl_args.mixing_alpha,
        )
        mix_hook = get_mix_data_hook(mixing_type=cl_args.mixing_type)

        init_kwargs["post_prepare_batch"].append(mix_hook)
        init_kwargs["loss"][0] = hook_mix_loss

    if len(cl_args.target_con_columns + cl_args.target_cat_columns) > 1:
        logger.debug(
            "Setting up hook for uncertainty weighted loss for multi task modelling."
        )
        uncertainty_hook = get_uncertainty_loss_hook(
            target_cat_columns=cl_args.target_cat_columns,
            target_con_columns=cl_args.target_con_columns,
            device=cl_args.device,
        )
        init_kwargs["loss"].append(uncertainty_hook)

    init_kwargs["loss"].append(hook_default_aggregate_losses)

    if cl_args.l1:
        init_kwargs["loss"].append(hook_add_l1_loss)

    return init_kwargs


@dataclass
class StepFunctionHookStages:

    al_hook = Callable[..., Dict]
    al_hooks = [Iterable[al_hook]]

    base_prepare_batch: al_hooks
    post_prepare_batch: al_hooks
    model_forward: al_hooks
    loss: al_hooks
    optimizer_backward: al_hooks
    metrics: al_hooks


def hook_default_prepare_batch(
    config: "Config",
    loader_batch: al_dataloader_getitem_batch,
    *args,
    **kwargs,
) -> Dict:

    batch = prepare_base_batch_default(
        loader_batch=loader_batch,
        cl_args=config.cl_args,
        target_columns=config.target_columns,
        model=config.model,
    )

    state_updates = {"batch": batch}

    return state_updates


def prepare_base_batch_default(
    loader_batch: al_dataloader_getitem_batch,
    cl_args: argparse.Namespace,
    target_columns: al_target_columns,
    model: fusion.FusionModel,
):

    inputs, target_labels, train_ids = loader_batch

    inputs_prepared = {}
    for input_name in inputs:

        if input_name.startswith("omics_"):
            cur_omics = inputs[input_name]
            cur_omics = cur_omics.to(device=cl_args.device)
            cur_omics = cur_omics.to(dtype=torch.float32)

            inputs_prepared[input_name] = cur_omics

        elif input_name.startswith("tabular_"):

            tabular_source_input: Dict[str, torch.Tensor] = inputs[input_name]
            for name, tensor in tabular_source_input.items():
                tabular_source_input[name] = tensor.to(device=cl_args.device)

            tabular = get_tabular_inputs(
                extra_cat_columns=cl_args.extra_cat_columns,
                extra_con_columns=cl_args.extra_con_columns,
                tabular_model=model.modules_to_fuse[input_name],
                tabular_input=tabular_source_input,
                device=cl_args.device,
            )
            inputs_prepared[input_name] = tabular
        else:
            raise ValueError(f"Unrecognized input type {input_name}.")

    if target_labels:
        target_labels = model_training_utils.parse_target_labels(
            target_columns=target_columns,
            device=cl_args.device,
            labels=target_labels,
        )

    batch = Batch(
        inputs=inputs_prepared,
        target_labels=target_labels,
        ids=train_ids,
    )

    return batch


def hook_default_model_forward(
    config: "Config", batch: "Batch", *args, **kwargs
) -> Dict:

    inputs = batch.inputs
    train_outputs = config.model(inputs=inputs)

    state_updates = {"model_outputs": train_outputs}

    return state_updates


def hook_default_optimizer_backward(
    config: "Config", state: Dict, *args, **kwargs
) -> Dict:

    optimizer_backward_kwargs = get_optimizer_backward_kwargs(
        optimizer_name=config.cl_args.optimizer
    )

    state["loss"].backward(**optimizer_backward_kwargs)
    config.optimizer.step()

    return {}


def hook_default_compute_metrics(
    config: "Config", batch: "Batch", state: Dict, *args, **kwargs
):

    train_batch_metrics = calculate_batch_metrics(
        target_columns=config.target_columns,
        outputs=state["model_outputs"],
        labels=batch.target_labels,
        mode="train",
        metric_record_dict=config.metrics,
    )

    train_batch_metrics_w_loss = add_loss_to_metrics(
        target_columns=config.target_columns,
        losses=state["per_target_train_losses"],
        metric_dict=train_batch_metrics,
    )

    train_batch_metrics_with_averages = add_multi_task_average_metrics(
        batch_metrics_dict=train_batch_metrics_w_loss,
        target_columns=config.target_columns,
        loss=state["loss"].item(),
        performance_average_functions=config.metrics["averaging_functions"],
    )

    state_updates = {"metrics": train_batch_metrics_with_averages}

    return state_updates


def hook_default_per_target_loss(
    config: "Config", batch: "Batch", state: Dict, *args, **kwargs
) -> Dict:

    per_target_train_losses = config.loss_function(
        inputs=state["model_outputs"], targets=batch.target_labels
    )

    state_updates = {"per_target_train_losses": per_target_train_losses}

    return state_updates


def hook_default_aggregate_losses(state: Dict, *args, **kwargs) -> Dict:

    train_loss_avg = aggregate_losses(losses_dict=state["per_target_train_losses"])
    state_updates = {"loss": train_loss_avg}

    return state_updates


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from functools import wraps, partial
from pathlib import Path
from typing import List, Tuple, Any, Union, Callable, Dict

import pandas as pd
import torch
from pandas import DataFrame


@dataclass
class ColumnOperation:
    """
    function: Function to apply on the label dataframe.
    """

    function: Callable[[DataFrame, str], DataFrame]
    function_args: Dict[str, Any]
    extra_columns_deps: Union[Tuple[str, ...], None] = ()
    only_apply_if_target: bool = False
    column_dtypes: Union[Dict[str, Any], None] = None


def streamline_df(
    df_func: Callable[[pd.DataFrame, Any], pd.DataFrame]
) -> Callable[[Any], pd.DataFrame]:
    @wraps(df_func)
    def wrapper(*args, df=None, column_name=None, **kwargs) -> pd.DataFrame:

        df = df_func(*args, df=df, column_name=column_name, **kwargs)

        return df

    return wrapper


@streamline_df
def placeholder_func(df, column_name):
    """
    Dummy function that we can pass to streamline_df so we still get it's functionality.
    """
    return df


def filter_ids_from_array_and_id_lists(
    ids_to_filter: List[str],
    id_list: List[str],
    arrays: Union[torch.Tensor, List[Path]],
    filter_type: str = "keep",
) -> Tuple[List[str], torch.Tensor]:
    """
    We use this function to perform filtering on the ID list and arrays, where these
    are assumed to be in the same order. This function is defined in the case where
    we want to dynamically filter out certain samples when setting up our dataloaders,
    e.g. we might want to fit on origin data with all "European" origin removed.

    :param ids_to_filter: Which IDs to filter on.
    :param id_list: A "master" list of IDs to perform the filtering operation on.
    :param arrays:  A "master" list of arrays to perform the filtering operation on.
    :param filter_type: Whether to keep or remove the IDs in `ids_to_filter`.
    :return: A tuple of (ids, arrays) after performing the filtering operation.
    """

    if filter_type not in ["keep", "skip"]:
        raise ValueError("Expecting either keep or skip for filter type.")

    keep_bool = filter_type == "keep"

    def condition(x):
        cond = x in ids_to_filter
        if keep_bool:
            return cond
        return not cond

    def cast_to_base_type(filtered_arrs):
        base_type = type(arrays)
        if base_type == torch.Tensor:
            return torch.stack(filtered_arrs)
        else:
            return base_type(filtered_arrs)

    filter_generator = (i for i in zip(id_list, arrays) if condition(i[0]))

    ids_filtered, arrays_filtered = zip(*filter_generator)

    return list(ids_filtered), cast_to_base_type(arrays_filtered)


def bucket_column(
    df: pd.DataFrame, column_name: str, n_buckets: int = 5, q_cut=False
) -> pd.DataFrame:
    """
    Splits continuous column into buckets for modelling with multi-class classification.

    :param df: Label dataframe in most cases.
    :param column_name: Column to bucket.
    :param n_buckets: Number of buckets.
    :param q_cut: Whether to do a quantile-based discretization with pd.qcut
    :return: Dataframe with ``column`` replaced with bucket-ized version.
    """
    cut_func = pd.qcut if q_cut else partial(pd.cut, include_lowest=True)
    df[column_name] = cut_func(df[column_name], n_buckets)

    return df


def get_low_count_ids(
    df: pd.DataFrame, column_name: str, threshold: int = 100
) -> List[str]:
    """
    Grabs those IDs where their counts are under a certain threshold. Why? We might
    want to restrict the number of classes during a multi-class experiment to only
    those that are above a certain threshold, i.e. it can happen that in 20k samples
    a certain class has <5 samples.

    Actually returns a ID list, so we can easily use it the output with
    `filter_ids_from_array_and_id_lists`.

    :param df: Dataframe to get counts on.
    :param column_name: Name of column to count on.
    :param threshold: Values under this threshold will be returned.
    :return: List of those IDs that have a value under the threshold.
    """
    counts = df[column_name].value_counts()
    low_count_ids = df[df.isin(counts.index[counts <= threshold]).values].index

    return list(low_count_ids)


def get_ids_from_values(
    df: pd.DataFrame, column_name: str, values_to_grab: List[Any]
) -> List[str]:
    """
    Grabs IDs matching certain values passed in. Why? We might want to manually filter
    out / keep some IDs matching a certain value in a given column.

    :param df: Dataframe to grab values from.
    :param column_name: Which columns to match values to.
    :param values_to_grab: Values to check in `transformer_name`.
    :return: List of IDs matching the values we're interested in.
    """

    ids = df[df[column_name].isin(values_to_grab)].index

    return list(ids)



from collections import OrderedDict
from dataclasses import dataclass
from typing import List, Tuple, TYPE_CHECKING, Union

import torch
from aislib import pytorch_utils
from aislib.misc_utils import get_logger
from aislib.pytorch_modules import Swish
from torch import nn

from eir.models.layers import FirstCNNBlock, SelfAttention, CNNResidualBlock

if TYPE_CHECKING:
    from eir.train import DataDimensions

logger = get_logger(__name__)


@dataclass
class CNNModelConfig:

    layers: Union[None, List[int]]

    fc_repr_dim: int

    down_stride: int
    first_stride_expansion: int

    channel_exp_base: int
    first_channel_expansion: int

    kernel_width: int
    first_kernel_expansion: int
    dilation_factor: int

    data_dimensions: "DataDimensions"

    rb_do: float

    sa: bool = False


class CNNModel(nn.Module):
    def __init__(self, model_config: CNNModelConfig):
        # TODO: Make work for heights, this means modifying stuff in layers.py
        super().__init__()

        self.model_config = model_config
        self.conv = nn.Sequential(*_make_conv_layers(self.resblocks, self.model_config))

        self.data_size_after_conv = pytorch_utils.calc_size_after_conv_sequence(
            self.model_config.data_dimensions.width, self.conv
        )
        self.no_out_channels = self.conv[-1].out_channels

        self.fc = nn.Sequential(
            OrderedDict(
                {
                    "fc_1_bn_1": nn.BatchNorm1d(self.fc_1_in_features),
                    "fc_1_act_1": Swish(),
                    "fc_1_linear_1": nn.Linear(
                        self.fc_1_in_features, self.model_config.fc_repr_dim, bias=False
                    ),
                }
            )
        )

        self._init_weights()

    @property
    def fc_1_in_features(self) -> int:
        return self.no_out_channels * self.data_size_after_conv

    @property
    def l1_penalized_weights(self) -> torch.Tensor:
        return self.conv[0].conv_1.weight

    @property
    def num_out_features(self) -> int:
        return self.model_config.fc_repr_dim

    def _init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                # Swish slope is roughly 0.5 around 0
                nn.init.kaiming_normal_(m.weight, a=0.5, mode="fan_out")
            elif isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.BatchNorm1d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    @property
    def resblocks(self) -> List[int]:
        if not self.model_config.layers:
            residual_blocks = find_no_cnn_resblocks_needed(
                self.model_config.data_dimensions.width,
                self.model_config.down_stride,
                self.model_config.first_stride_expansion,
            )
            logger.info(
                "No residual blocks specified in CL args, using input "
                "%s based on width approximation calculation.",
                residual_blocks,
            )
            return residual_blocks
        return self.model_config.layers

    def forward(self, input: torch.Tensor) -> torch.Tensor:

        out = self.conv(input)
        out = out.view(out.shape[0], -1)

        out = self.fc(out)

        return out


def _make_conv_layers(
    residual_blocks: List[int], cnn_model_configuration: CNNModelConfig
) -> List[nn.Module]:
    """
    Used to set up the convolutional layers for the model. Based on the passed in
    residual blocks, we want to set up the actual blocks with all the relevant
    convolution parameters.

    We start with a base channel number of 2**5 == 32.

    Also inserts a self-attention layer in just before the last residual block.

    :param residual_blocks: List of ints, where each int indicates number of blocks w.
    that channel dimension.
    :param cnn_model_configuration: Experiment hyperparameters / configuration needed
    for the convolution setup.
    :return: A list of `nn.Module` objects to be passed to `nn.Sequential`.
    """
    mc = cnn_model_configuration

    down_stride_w = mc.down_stride

    first_conv_channels = 2 ** mc.channel_exp_base * mc.first_channel_expansion
    first_conv_kernel = mc.kernel_width * mc.first_kernel_expansion
    first_conv_stride = down_stride_w * mc.first_stride_expansion

    first_kernel, first_pad = pytorch_utils.calc_conv_params_needed(
        input_width=mc.data_dimensions.width,
        kernel_size=first_conv_kernel,
        stride=first_conv_stride,
        dilation=1,
    )

    conv_blocks = [
        FirstCNNBlock(
            in_channels=cnn_model_configuration.data_dimensions.channels,
            out_channels=first_conv_channels,
            conv_1_kernel_h=cnn_model_configuration.data_dimensions.height,
            conv_1_kernel_w=first_kernel,
            conv_1_padding=first_pad,
            down_stride_w=first_conv_stride,
            dilation=1,
            rb_do=mc.rb_do,
        )
    ]

    sa_added = False
    for layer_arch_idx, layer_arch_layers in enumerate(residual_blocks):
        for layer in range(layer_arch_layers):
            cur_layer, cur_width = _get_conv_resblock(
                conv_blocks=conv_blocks,
                layer_arch_idx=layer_arch_idx,
                down_stride=down_stride_w,
                cnn_config=cnn_model_configuration,
            )

            if mc.sa and cur_width < 1024 and not sa_added:
                attention_channels = conv_blocks[-1].out_channels
                conv_blocks.append(SelfAttention(attention_channels))
                sa_added = True

            conv_blocks.append(cur_layer)

    return conv_blocks


def _get_conv_resblock(
    conv_blocks: List[nn.Module],
    layer_arch_idx: int,
    down_stride: int,
    cnn_config: CNNModelConfig,
) -> Tuple[CNNResidualBlock, int]:
    mc = cnn_config

    cur_conv = nn.Sequential(*conv_blocks)
    cur_width = pytorch_utils.calc_size_after_conv_sequence(
        input_width=mc.data_dimensions.width, conv_sequence=cur_conv
    )

    cur_kern, cur_padd = pytorch_utils.calc_conv_params_needed(
        input_width=cur_width,
        kernel_size=mc.kernel_width,
        stride=down_stride,
        dilation=1,
    )

    cur_block_number = (
        len([i for i in conv_blocks if isinstance(i, CNNResidualBlock)]) + 1
    )
    cur_dilation_factor = _get_cur_dilation(
        dilation_factor=mc.dilation_factor,
        width=cur_width,
        block_number=cur_block_number,
    )

    cur_in_channels = conv_blocks[-1].out_channels
    cur_out_channels = 2 ** (mc.channel_exp_base + layer_arch_idx)

    cur_layer = CNNResidualBlock(
        in_channels=cur_in_channels,
        out_channels=cur_out_channels,
        conv_1_kernel_h=1,
        conv_1_kernel_w=cur_kern,
        conv_1_padding=cur_padd,
        down_stride_w=down_stride,
        dilation=cur_dilation_factor,
        full_preact=True if len(conv_blocks) == 1 else False,
        rb_do=mc.rb_do,
    )

    return cur_layer, cur_width


def _get_cur_dilation(dilation_factor: int, width: int, block_number: int):
    """
    Note that block_number refers to the number of residual blocks (not first block
    or self attention).
    """
    dilation = dilation_factor ** block_number

    while dilation >= width:
        dilation = dilation // dilation_factor

    return dilation


def find_no_cnn_resblocks_needed(
    width: int, stride: int, first_stride_expansion: int
) -> List[int]:
    """
    Used in order to calculate / set up residual blocks specifications as a list
    automatically when they are not passed in as CL args, based on the minimum
    width after the resblock convolutions.

    We have 2 resblocks per channel depth until we have a total of 8 blocks,
    then the rest is put in the third depth index (following resnet convention).

    That is with a base channel depth of 32, we have these depths in the list:
    [32, 64, 128, 256].

    Examples
    ------
    3 blocks --> [2, 1]
    7 blocks --> [2, 2, 2, 1]
    10 blocks --> [2, 2, 4, 2]
    """

    min_size = 8 * stride
    # account for first conv
    cur_width = width // (stride * first_stride_expansion)

    resblocks = [0] * 4
    while cur_width >= min_size:
        cur_no_blocks = sum(resblocks)

        if cur_no_blocks >= 8:
            resblocks[2] += 1
        else:
            cur_index = cur_no_blocks // 2
            resblocks[cur_index] += 1

        cur_width = cur_width // stride

    return [i for i in resblocks if i != 0]



import csv
from copy import copy
import warnings
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Dict, TYPE_CHECKING, List, Tuple, Callable, Union

import numpy as np
import pandas as pd
import torch
from aislib.misc_utils import ensure_path_exists, get_logger
from scipy.stats import pearsonr
from sklearn.metrics import (
    matthews_corrcoef,
    r2_score,
    mean_squared_error,
    roc_auc_score,
    average_precision_score,
    accuracy_score,
)
from sklearn.preprocessing import StandardScaler, label_binarize
from torch import nn
from torch.utils.tensorboard import SummaryWriter

from eir.data_load.data_utils import get_target_columns_generator
from eir.data_load.label_setup import al_label_transformers

if TYPE_CHECKING:
    from eir.train import al_criterions, Config  # noqa: F401
    from eir.models.omics.omics_models import al_models  # noqa: F401
    from eir.train_utils.train_handlers import HandlerConfig
    from eir.data_load.label_setup import (  # noqa: F401
        al_target_columns,
        al_label_transformers_object,
    )

# aliases
al_step_metric_dict = Dict[str, Dict[str, float]]
al_metric_record_dict = Dict[
    str, Union[Tuple["MetricRecord", ...], "al_averaging_functions_dict"]
]
al_averaging_functions_dict = Dict[
    str, Callable[["al_step_metric_dict", str, str], float]
]

logger = get_logger(name=__name__, tqdm_compatible=True)


@dataclass()
class MetricRecord:
    name: str
    function: Callable
    only_val: bool = False
    minimize_goal: bool = False


def calculate_batch_metrics(
    target_columns: "al_target_columns",
    outputs: Dict[str, torch.Tensor],
    labels: Dict[str, torch.Tensor],
    mode: str,
    metric_record_dict: al_metric_record_dict,
) -> al_step_metric_dict:
    assert mode in ["val", "train"]

    target_columns_gen = get_target_columns_generator(target_columns=target_columns)

    master_metric_dict = {}

    for column_type, column_name in target_columns_gen:
        cur_metric_dict = {}

        cur_metric_records: Tuple[MetricRecord, ...] = metric_record_dict[column_type]
        cur_outputs = outputs[column_name].detach().cpu().numpy()
        cur_labels = labels[column_name].cpu().numpy()

        for metric_record in cur_metric_records:

            if metric_record.only_val and mode == "train":
                continue

            cur_key = f"{column_name}_{metric_record.name}"
            cur_metric_dict[cur_key] = metric_record.function(
                outputs=cur_outputs, labels=cur_labels, column_name=column_name
            )

        master_metric_dict[column_name] = cur_metric_dict

    return master_metric_dict


def add_loss_to_metrics(
    target_columns: "al_target_columns",
    losses: Dict[str, torch.Tensor],
    metric_dict: al_step_metric_dict,
) -> al_step_metric_dict:

    target_columns_gen = get_target_columns_generator(target_columns=target_columns)
    metric_dict_copy = copy(metric_dict)

    for column_type, column_name in target_columns_gen:
        cur_metric_dict = metric_dict_copy[column_name]
        cur_metric_dict[f"{column_name}_loss"] = losses[column_name].item()

    return metric_dict_copy


def add_multi_task_average_metrics(
    batch_metrics_dict: al_step_metric_dict,
    target_columns: "al_target_columns",
    loss: float,
    performance_average_functions: Dict[str, Callable[[al_step_metric_dict], float]],
) -> al_step_metric_dict:
    average_performance = average_performances_across_tasks(
        metric_dict=batch_metrics_dict,
        target_columns=target_columns,
        performance_calculation_functions=performance_average_functions,
    )
    batch_metrics_dict["average"] = {
        "loss-average": loss,
        "perf-average": average_performance,
    }

    return batch_metrics_dict


def average_performances_across_tasks(
    metric_dict: al_step_metric_dict,
    target_columns: "al_target_columns",
    performance_calculation_functions: Dict[
        str, Callable[[al_step_metric_dict], float]
    ],
) -> float:
    target_columns_gen = get_target_columns_generator(target_columns)

    all_metrics = []

    for column_type, column_name in target_columns_gen:
        cur_metric_func = performance_calculation_functions.get(column_type)

        metric_func_args = {"metric_dict": metric_dict, "column_name": column_name}
        cur_value = cur_metric_func(**metric_func_args)
        all_metrics.append(cur_value)

        all_metrics.append(cur_value)

    average = np.array(all_metrics).mean()

    return average


def calc_mcc(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    pred = np.argmax(a=outputs, axis=1)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        mcc = matthews_corrcoef(labels, pred)

    return mcc


def calc_roc_auc_ovr(
    outputs: np.ndarray, labels: np.ndarray, average: str = "macro", *args, **kwargs
) -> float:
    """
    TODO:   In rare scenarios, we might run into the issue of not having all labels
            represented in the labels array (i.e. labels were in train, but not in
            valid). This is not a problem for metrics like MCC / accuracy, but we
            will have to account for this here and in the AP calculation, possibly
            by ignoring columns in outputs and label_binarize outputs where the columns
            returned from label_binarize are all 0.
    """

    assert average in ["micro", "macro"]

    if outputs.shape[1] > 2:
        labels = label_binarize(y=labels, classes=sorted(np.unique(labels)))
    else:
        outputs = outputs[:, 1]

    roc_auc = roc_auc_score(y_true=labels, y_score=outputs, average=average)
    return roc_auc


def calc_average_precision_ovr(
    outputs: np.ndarray, labels: np.ndarray, average: str = "macro", *args, **kwargs
) -> float:

    assert average in ["micro", "macro"]

    if outputs.shape[1] > 2:
        labels = label_binarize(y=labels, classes=sorted(np.unique(labels)))
    else:
        outputs = outputs[:, 1]

    average_precision = average_precision_score(
        y_true=labels, y_score=outputs, average=average
    )

    return average_precision


def calc_acc(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:
    pred = np.argmax(outputs, axis=1)

    accuracy = accuracy_score(y_true=labels, y_pred=pred)
    return accuracy


def calc_pcc(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:

    if len(outputs) < 2:
        return 0.0

    pcc = pearsonr(x=labels.squeeze(), y=outputs.squeeze())[0]
    return pcc


def calc_r2(outputs: np.ndarray, labels: np.ndarray, *args, **kwargs) -> float:

    if len(outputs) < 2:
        return 0.0

    r2 = r2_score(y_true=labels.squeeze(), y_pred=outputs.squeeze())
    return r2


def calc_rmse(
    outputs: torch.Tensor,
    labels: torch.Tensor,
    target_transformers: Dict[str, StandardScaler],
    column_name: str,
    *args,
    **kwargs,
) -> float:
    cur_target_transformer = target_transformers[column_name]

    labels = cur_target_transformer.inverse_transform(labels).squeeze()
    preds = cur_target_transformer.inverse_transform(outputs).squeeze()

    rmse = np.sqrt(mean_squared_error(y_true=labels, y_pred=preds))
    return rmse


def calculate_prediction_losses(
    criterions: "al_criterions",
    inputs: Dict[str, torch.Tensor],
    targets: Dict[str, torch.Tensor],
) -> Dict[str, torch.Tensor]:
    losses_dict = {}

    for target_column, criterion in criterions.items():
        cur_target_col_labels = targets[target_column]
        cur_target_col_outputs = inputs[target_column]
        losses_dict[target_column] = criterion(
            input=cur_target_col_outputs, target=cur_target_col_labels
        )

    return losses_dict


def aggregate_losses(losses_dict: Dict[str, torch.Tensor]) -> torch.Tensor:
    losses_values = list(losses_dict.values())
    average_loss = torch.mean(torch.stack(losses_values))

    return average_loss


def get_uncertainty_loss_hook(
    target_cat_columns: List[str],
    target_con_columns: List[str],
    device: str,
):
    loss_module = UncertaintyMultiTaskLoss(
        target_cat_columns=target_cat_columns,
        target_con_columns=target_con_columns,
        device=device,
    )

    hook = partial(hook_add_uncertainty_loss, uncertainty_module=loss_module)

    return hook


def hook_add_uncertainty_loss(
    state: Dict,
    uncertainty_module: "UncertaintyMultiTaskLoss",
    loss_key: str = "per_target_train_losses",
    *args,
    **kwargs,
):
    base_losses_dict = state[loss_key]

    uncertainty_losses = uncertainty_module(losses_dict=base_losses_dict)

    state_updates = {loss_key: uncertainty_losses}

    return state_updates


class UncertaintyMultiTaskLoss(nn.Module):
    def __init__(
        self,
        target_cat_columns: List[str],
        target_con_columns: List[str],
        device: str,
    ):
        super().__init__()

        self.target_cat_columns = target_cat_columns
        self.target_con_columns = target_con_columns
        self.device = device

        self.log_vars = self._construct_params(
            cur_target_columns=self.target_cat_columns + self.target_con_columns,
            device=self.device,
        )

    @staticmethod
    def _construct_params(cur_target_columns: List[str], device: str):
        param_dict = {}
        for column_name in cur_target_columns:
            param_dict[column_name] = nn.Parameter(
                torch.zeros(1), requires_grad=True
            ).to(device=device)

        return param_dict

    def _calc_uncertainty_loss(self, name, loss_value):
        log_var = self.log_vars[name]
        scalar = 2.0 if name in self.target_cat_columns else 1.0

        precision = torch.exp(-log_var)
        loss = scalar * torch.sum(precision * loss_value + log_var)

        return loss

    def forward(self, losses_dict: Dict):
        losses_uncertain = {}

        for target_column, loss_value_base in losses_dict.items():
            loss_value_uncertain = self._calc_uncertainty_loss(
                name=target_column, loss_value=loss_value_base
            )
            losses_uncertain[target_column] = loss_value_uncertain

        return losses_uncertain


def hook_add_l1_loss(
    config: "Config",
    state: Dict,
    loss_key: str = "loss",
    *args,
    **kwargs,
) -> Dict:
    l1_loss = get_model_l1_loss(model=config.model, l1_weight=config.cl_args.l1)

    updated_loss = state[loss_key] + l1_loss

    state_updates = {loss_key: updated_loss}

    return state_updates


def get_model_l1_loss(model: "al_models", l1_weight: float) -> torch.Tensor:
    l1_loss = calc_l1_loss(
        weight_tensor=model.l1_penalized_weights, l1_weight=l1_weight
    )
    return l1_loss


def calc_l1_loss(weight_tensor: torch.Tensor, l1_weight: float):
    l1_loss = torch.norm(weight_tensor, p=1) * l1_weight
    return l1_loss


def add_extra_losses(total_loss: torch.Tensor, extra_loss_functions: List[Callable]):
    """
    TODO: Possibly add inputs and labels as arguments here if needed later.
    """
    for loss_func in extra_loss_functions:
        total_loss += loss_func()

    return total_loss


def persist_metrics(
    handler_config: "HandlerConfig",
    metrics_dict: "al_step_metric_dict",
    iteration: int,
    write_header: bool,
    prefixes: Dict[str, str],
):

    hc = handler_config
    c = handler_config.config
    cl_args = c.cl_args

    metrics_files = get_metrics_files(
        target_columns=c.target_columns,
        run_folder=hc.run_folder,
        train_or_val_target_prefix=f"{prefixes['metrics']}",
    )

    if write_header:
        _ensure_metrics_paths_exists(metrics_files=metrics_files)

    for metrics_name, metrics_history_file in metrics_files.items():
        cur_metric_dict = metrics_dict[metrics_name]

        _add_metrics_to_writer(
            name=f"{prefixes['writer']}/{metrics_name}",
            metric_dict=cur_metric_dict,
            iteration=iteration,
            writer=c.writer,
            plot_skip_steps=cl_args.plot_skip_steps,
        )

        _append_metrics_to_file(
            filepath=metrics_history_file,
            metrics=cur_metric_dict,
            iteration=iteration,
            write_header=write_header,
        )


def get_metrics_files(
    target_columns: "al_target_columns",
    run_folder: Path,
    train_or_val_target_prefix: str,
) -> Dict[str, Path]:
    assert train_or_val_target_prefix in ["validation_", "train_"]

    all_target_columns = target_columns["con"] + target_columns["cat"]

    path_dict = {}
    for target_column in all_target_columns:
        cur_fname = train_or_val_target_prefix + target_column + "_history.log"
        cur_path = Path(run_folder, "results", target_column, cur_fname)
        path_dict[target_column] = cur_path

    average_loss_training_metrics_file = get_average_history_filepath(
        run_folder=run_folder, train_or_val_target_prefix=train_or_val_target_prefix
    )
    path_dict["average"] = average_loss_training_metrics_file

    return path_dict


def get_average_history_filepath(
    run_folder: Path, train_or_val_target_prefix: str
) -> Path:
    assert train_or_val_target_prefix in ["validation_", "train_"]
    metrics_file_path = run_folder / f"{train_or_val_target_prefix}average_history.log"
    return metrics_file_path


def _ensure_metrics_paths_exists(metrics_files: Dict[str, Path]) -> None:
    for path in metrics_files.values():
        ensure_path_exists(path)


def _add_metrics_to_writer(
    name: str,
    metric_dict: Dict[str, float],
    iteration: int,
    writer: SummaryWriter,
    plot_skip_steps: int,
) -> None:
    """
    We do %10 to reduce the amount of training data going to tensorboard, otherwise
    it slows down with many large experiments.
    """
    if iteration >= plot_skip_steps and iteration % 10 == 0:
        for metric_name, metric_value in metric_dict.items():
            cur_name = name + f"/{metric_name}"
            writer.add_scalar(
                tag=cur_name, scalar_value=metric_value, global_step=iteration
            )


def _append_metrics_to_file(
    filepath: Path, metrics: Dict[str, float], iteration: int, write_header=False
):
    """
    TODO:   Have cached file handles here instead of reopening the file at every
            iteration.
    """
    with open(str(filepath), "a") as logfile:
        fieldnames = ["iteration"] + sorted(metrics.keys())
        writer = csv.DictWriter(logfile, fieldnames=fieldnames)

        if write_header:
            writer.writeheader()

        dict_to_write = {**{"iteration": iteration}, **metrics}
        writer.writerow(dict_to_write)


def read_metrics_history_file(file_path: Path) -> pd.DataFrame:
    df = pd.read_csv(file_path, index_col="iteration")

    return df


def get_metrics_dataframes(
    results_dir: Path, target_string: str
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    train_history_path = read_metrics_history_file(
        results_dir / f"train_{target_string}_history.log"
    )
    valid_history_path = read_metrics_history_file(
        results_dir / f"validation_{target_string}_history.log"
    )

    return train_history_path, valid_history_path


def get_default_metrics(
    target_transformers: al_label_transformers,
) -> "al_metric_record_dict":
    mcc = MetricRecord(name="mcc", function=calc_mcc)
    acc = MetricRecord(name="acc", function=calc_acc)

    rmse = MetricRecord(
        name="rmse",
        function=partial(calc_rmse, target_transformers=target_transformers),
        minimize_goal=True,
    )

    roc_auc_macro = MetricRecord(
        name="roc-auc-macro", function=calc_roc_auc_ovr, only_val=True
    )
    ap_macro = MetricRecord(
        name="ap-macro", function=calc_average_precision_ovr, only_val=True
    )
    r2 = MetricRecord(name="r2", function=calc_r2, only_val=True)
    pcc = MetricRecord(name="pcc", function=calc_pcc, only_val=True)

    averaging_functions = get_default_performance_averaging_functions(
        cat_metric_name="mcc", con_metric_name="loss"
    )
    default_metrics = {
        "cat": (mcc, acc, roc_auc_macro, ap_macro),
        "con": (rmse, r2, pcc),
        "averaging_functions": averaging_functions,
    }
    return default_metrics


def get_default_performance_averaging_functions(
    cat_metric_name: str, con_metric_name: str
) -> al_averaging_functions_dict:
    def _calc_cat_averaging_value(
        metric_dict: "al_step_metric_dict", column_name: str, metric_name: str
    ) -> float:
        return metric_dict[column_name][f"{column_name}_{metric_name}"]

    def _calc_con_averaging_value(
        metric_dict: "al_step_metric_dict", column_name: str, metric_name: str
    ) -> float:
        return 1 - metric_dict[column_name][f"{column_name}_{metric_name}"]

    performance_averaging_functions = {
        "cat": partial(_calc_cat_averaging_value, metric_name=cat_metric_name),
        "con": partial(_calc_con_averaging_value, metric_name=con_metric_name),
    }

    return performance_averaging_functions



from collections import Counter
from typing import TYPE_CHECKING, List, Tuple, Union, Dict, Iterable

import torch
from aislib.misc_utils import get_logger
from torch.utils.data import WeightedRandomSampler

if TYPE_CHECKING:
    from eir.data_load.datasets import (  # noqa: F401
        DatasetBase,
        Sample,
    )

logger = get_logger(name=__name__, tqdm_compatible=True)


# Type Aliases
al_sample_weight_and_counts = Dict[str, Union[torch.Tensor, List[int]]]


def get_weighted_random_sampler(samples: Iterable["Sample"], target_columns: List[str]):
    """
    Labels spec:

    {
        {
        ID1:
            {
                Label Column: Target Value,
                Extra Column 1: Extra Column 1 Value
                Extra Column 2: Extra Column 2 Value}
            }
        },
        ID2: {...}
    }

    The list comprehension is going over all the label dicts associated with the IDs,
    then just parsing the label (converting to int in the case of classification).
    """
    all_column_weights = _gather_column_sampling_weights(
        samples=samples, target_columns=target_columns
    )

    samples_weighted, num_sample_per_epoch = _aggregate_column_sampling_weights(
        all_target_columns_weights_and_counts=all_column_weights
    )

    logger.debug("Num samples per epoch: %d", num_sample_per_epoch)
    sampler = WeightedRandomSampler(
        weights=samples_weighted, num_samples=num_sample_per_epoch, replacement=True
    )

    return sampler


def _gather_column_sampling_weights(
    samples: Iterable["Sample"], target_columns: Iterable[str]
) -> Dict[str, al_sample_weight_and_counts]:
    all_target_label_weight_dicts = {}

    for column in target_columns:
        cur_label_iterable = (i.target_labels[column] for i in samples)
        cur_weight_dict = _get_column_label_weights_and_counts(
            label_iterable=cur_label_iterable
        )

        logger.debug(
            "Label counts in column %s:  %s", column, cur_weight_dict["label_counts"]
        )

        all_target_label_weight_dicts[column] = cur_weight_dict

    return all_target_label_weight_dicts


def _get_column_label_weights_and_counts(
    label_iterable: Iterable[int],
) -> al_sample_weight_and_counts:
    """
    We have the assertion to make sure we have a unique integer for each label, starting
    with 0 as we use it to index into the weights directly.

    TODO:   Optimize so we do just one pass over `train_dataset.samples` if this becomes
            a bottleneck.
    """

    def _check_labels(label_list: List[int]):
        labels_set = set(label_list)
        assert sorted(list(labels_set)) == list(range(len(labels_set)))

    labels = list(label_iterable)
    _check_labels(label_list=labels)

    label_counts = [i[1] for i in sorted(Counter(labels).items())]

    weights = 1.0 / torch.tensor(label_counts, dtype=torch.float32)
    samples_weighted = weights[labels]

    output_dict = {"samples_weighted": samples_weighted, "label_counts": label_counts}
    return output_dict


def _aggregate_column_sampling_weights(
    all_target_columns_weights_and_counts: Dict[str, al_sample_weight_and_counts]
) -> Tuple[torch.Tensor, int]:
    """
    We sum up the normalized weights for each target column to create the final sampling
    weights.

    As for the samples per epoch, we want to take the sum of the class with the fewest
    counts over all target columns to sample on.
    """

    all_weights = torch.stack(
        [i["samples_weighted"] for i in all_target_columns_weights_and_counts.values()],
        dim=1,
    )
    all_weights_summed = all_weights.sum(dim=1)

    samples_per_epoch = sum(
        min(i["label_counts"]) for i in all_target_columns_weights_and_counts.values()
    )
    samples_per_epoch = min(len(all_weights_summed), samples_per_epoch)

    return all_weights_summed, samples_per_epoch
